#!/usr/bin/env python
# filename: cluster.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os

# import platform
import random
import string
import subprocess as sp

# import sys
import tempfile
from typing import Iterable, Optional, Union

# from Bio.Align import AlignInfo
from ..bin import get_path as get_binary_path
from ..core.pair import Pair
from ..core.sequence import Sequence
from ..io import read_fasta, to_fasta

# from ..utils.decorators import lazy_property
# from ..utils.pipeline import make_dir
from .alignment import make_consensus

__all__ = [
    "cluster",
    "cluster_vsearch",
    "cluster_mmseqs",
    "cluster_cdhit",
    "Clusters",
    "Cluster",
]


class Cluster:
    """
    Class for a sequence cluster.

    Attributes
    ----------
    name : str
        Name of the cluster.

    sequences : list
        List of sequences in the cluster.

    size : int
        Number of sequences in the cluster.

    seq_ids : list
        List of sequence IDs in the cluster.

    centroid : ``Sequence``
        Centroid sequence of the cluster.

    consensus : ``Sequence``
        Consensus sequence of the cluster.

    """

    def __init__(self, name, sequences, centroid=None):
        self.name = name
        self.sequences = sequences
        self.centroid = centroid
        self._consensus = None

    def __iter__(self):
        for s in self.sequences:
            yield s

    @property
    def size(self) -> int:
        return len(self.sequences)

    @property
    def sequence_ids(self) -> Iterable[str]:
        return [s.id for s in self.sequences]

    @property
    def seq_ids(self) -> Iterable[str]:
        # retained for backwards compatibility, use sequence_ids instead
        return [s.id for s in self.sequences]

    @property
    def consensus(self) -> Sequence:
        if self._consensus is None:
            self.make_consensus()
        return self._consensus()

    def make_consensus(
        self,
        threshold: float = 0.51,
        algo: str = "mafft",
        ambiguous: Optional[str] = None,
    ) -> Sequence:
        """
        Makes a consensus sequence from the cluster's sequences.

        Parameters
        ----------
        threshold : float, default=0.51
            Threshold for calling a consensus base. Must be between 0 and 1.

        algo : str, default="mafft"
            Algorithm to be used for generating the consensus sequence. Options are
            ``"mafft"``, ``"famsa"``, or ``"muscle"``.

        ambiguous : str, default=None
            Character to use for ambiguous bases. If not provided, the default
            ambiguous base for the sequence type will be used.

        Returns
        -------
        consensus : ``Sequence``

        """
        if len(self.sequences) == 1:
            self._consensus = self.sequences[0]
        else:
            consensus = make_consensus(
                self.sequences,
                threshold=threshold,
                algo=algo,
                ambiguous=ambiguous,
            )
            self._consensus = consensus
        return self._consensus
        # return consensus
        # aln = mafft(self.sequences)
        # if aln is None:
        #     print("ERROR: Failed to generate an alignmnet for a consensus sequence.")
        #     return None
        # summary_align = AlignInfo.SummaryInfo(aln)
        # consensus = summary_align.gap_consensus(threshold=0.51, ambiguous="n")
        # consensus_string = str(consensus).replace("-", "")
        # consensus_seq = Sequence(consensus_string.upper())
        # return consensus_seq


class Clusters:
    """
    Class for a collection of clusters.

    Attributes
    ----------
    clusters : list
        List of clusters. Clusters are sorted in descending order by size,
        with the largest cluster first.

    centroids : list
        List of cluster centroids.

    largest_cluster : ``Cluster``
        The largest cluster in the collection.

    count : int
        Number of clusters in the collection.


    """

    def __init__(self, clusters=None):
        self._clusters = self._parse_clusters(clusters)

    def __iter__(self):
        for cluster in self.clusters:
            yield cluster

    def __getitem__(self, key):
        return self.clusters[key]

    def __len__(self):
        return len(self._clusters)

    @property
    def clusters(self):
        return sorted(self._clusters, key=lambda x: x.size, reverse=True)

    @property
    def centroids(self):
        return [c.centroid for c in self.clusters]

    @property
    def largest_cluster(self):
        return self.clusters[0]

    @property
    def count(self):
        return len(self.clusters)

    def _parse_clusters(self, clusts):
        if isinstance(clusts, dict):
            clusters = []
            for name, cdata in clusts.items():
                seqs = cdata["seqs"]
                centroid = cdata["centroid"]
                clusters.append(Cluster(name, seqs, centroid=centroid))
            return clusters
        elif clusts is None:
            return []
        else:
            return clusts

    def add(self, cluster):
        self._clusters.append(cluster)


def cluster(
    sequences: Union[Iterable, str],
    threshold: float = 0.975,
    algo: str = "auto",
    temp_dir: str = "/tmp",
    iddef: int = 0,
    cluster_mode: str = "2",
    cov_mode: str = "0",
    coverage: float = 0.8,
    alignment_mode: str = "3",
    seq_id_mode: str = "1",
    vsearch_bin: str = None,
    mmseqs_bin: str = None,
    cdhit_bin: str = None,
    id_key: Optional[str] = None,
    seq_key: Optional[str] = None,
    threads: Optional[int] = None,
    strand: str = "plus",
    as_dict: bool = False,
    debug: bool = False,
) -> Union[dict, Clusters]:
    """
    Clusters sequences using `CD-HIT`_, `VSEARCH`_ or `MMseqs2`_. By default, sequences will
    be clustered with VSEARCH if there are fewer than 10,000 nucleotide sequences, with
    CD-HIT if there are fewer than 10,000 amino acide sequences, and with MMseqs2 if there
    are more than 10,000 sequences (nucleotide or amino acid). These defaults can be
    overridden with `algo`.

    Parameters
    ----------
    sequences : iterable or string
        Input sequences in any of the following formats:
            1. list of abutils ``Sequence`` objects
            2. FASTA-formatted string
            3. path to a FASTA-formatted file
            4. list of BioPython ``SeqRecord`` objects
            5. list of lists/tuples, of the format ``[sequence_id, sequence]``
            6. list of strings, with each string being a separate sequence.
        Required.

    .. note::

        If ``sequences`` is a list of ``str`` objects, they will be assigned random IDs.

    threshold : float, default=0.975
        Identity threshold for clustering. Must be between 0 and 1.

    algo : float, default="auto"
        Algorithm to be used for clustering. Options are ``"vsearch"``, ``"mmseqs"``, ``"cdhit"``,
        or ``"auto"``. By default (``"auto"``), VSEARCH will be used if there are fewer than 10,000
        nucleotide sequences, CD-HIT will be used if there are fewer than 10,000 amino acid sequences,
        and MMseqs2 will be used for 10,000 sequences or more. Providing ``"vsearch"``, ``"cdhit"``,
        or ``"mmseqs"`` will force the use of the desired clustering algorithm regardless of the
        number or sequences to be clustered.

    temp_dir : str, default="/tmp"
        Path to a directory for temporary storage of clustering files.

    iddef : int, default=1
        Identity definition, as implemented by VSEARCH. Options are:
            0. CD-HIT definition: (matching columns) / (shortest sequence length).
            1. edit distance: (matching columns) / (alignment length).
            2. edit distance excluding terminal gaps (same as --id).
            3. Marine Biological Lab definition counting each gap opening (internal or
               terminal) as a single mismatch, whether or not the gap was extended: 1.0
               - [(mismatches + gap openings)/(longest sequence length)]
            4. BLAST definition, equivalent to --iddef 1 in a context of global pairwise
               alignment.

    cluster_mode : str, default="2"
        Clustering mode, as implemented by MMseqs2. Options are:
            1. greedy set cover
            2. connected component
            3. greedy incremental (CD-HIT-like)

    cov_mode : str, default="0"
        Coverage mode, as implemented by MMseqs2. Options are:
            0. bidirectional
            1. target coverage
            2. query coverage
            3. target-in-query length coverage

    coverage : float, default=0.8
        Coverage threshold for clustering with MMseqs2. Must be between 0 and 1.

    alignment_mode : str, default="3"
        Alignment mode, as implemented by MMseqs2. Options are:
            0. automatic
            1. only score and end_pos
            2. also start_pos and cov
            3. also seq.id
            4. only ungapped alignment

    seq_id_mode : str, default="1"
        Sequence ID mode, as implemented by MMseqs2. Options are:
            0. alignment length
            1. shorter sequence length
            2. longer sequence length

    vsearch_bin : str, optional
        Path to a VSEARCH executable. If not provided, the VSEARCH binary bundled
        with ``abutils`` will be used.

    mmseqs_bin : str, optional
        Path to a MMseqs2 executable. If not provided, the MMseqs2 binary bundled
        with ``abutils`` will be used.

    cdhit_bin : str, optional
        Path to a CD-HIT executable. If not provided, the CD-HIT binary bundled
        with ``abutils`` will be used.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    strand : str, default="plus"
        Strand of the sequences to align. Options are ``"plus"`` and ``"both"``.

    as_dict : bool, default=False
        If ``True``, return clustering results as a ``dict`` rather than a ``Clusters``
        object. the ``dict`` is of the format:
            {"cluster1_name": {"centroid": cluster1_centroid,
                               "seqs": [seq1, seq2, seq3, ...]},
             "cluster2_name": {"centroid": cluster2_centroid,
                               "seqs": [seq4, seq5, seq6, ...]},
            }

    debug : bool, default=False
        If ``True``, prints MAFFT's standard output and standard error.
        Default is ``False``.


    Returns
    -------
    clusters : ``Clusters`` or ``dict``


    .. _VSEARCH
        https://github.com/torognes/vsearch
    .. _MMseqs2
        https://github.com/soedinglab/MMseqs2
    .. _CD-HIT
       https://github.com/weizhongli/cdhit/wiki

    """
    # check input data to get the number of sequences
    fasta_file = to_fasta(
        sequences, tempfile_dir=temp_dir, id_key=id_key, sequence_key=seq_key
    )
    seqs = read_fasta(fasta_file)
    # check for AA sequences (which can't be handled by USEARCH)
    nt_residues = set(["A", "C", "G", "T", "N", "U", "-"])
    is_aa = False
    for s in seqs[:25]:
        if any([r not in nt_residues for r in s.sequence]):
            is_aa = True
    seq_dict = {s.id: s for s in seqs}
    # select the clustering algo
    algo = algo.lower()
    if algo == "auto" and len(seqs) < 10000:
        if is_aa:
            algo = "cdhit"
        else:
            algo = "vsearch"
    elif algo == "auto" and len(seqs) >= 10000:
        algo = "mmseqs"
    if algo in ["mmseqs", "mmseqs2"]:
        cluster_dict = cluster_mmseqs(
            fasta_file=fasta_file,
            threshold=threshold,
            cluster_mode=cluster_mode,
            cov_mode=cov_mode,
            coverage=coverage,
            alignment_mode=alignment_mode,
            seq_id_mode=seq_id_mode,
            threads=threads,
            temp_dir=temp_dir,
            mmseqs_bin=mmseqs_bin,
            as_dict=True,
            debug=debug,
        )
    elif algo in ["cdhit", "cd_hit", "cd-hit"]:
        cluster_dict = cluster_cdhit(
            fasta_file=fasta_file,
            threshold=threshold,
            temp_dir=temp_dir,
            cdhit_bin=cdhit_bin,
            threads=0 if threads is None else threads,
            as_dict=True,
            debug=debug,
        )
    elif algo == "vsearch":
        cluster_dict = cluster_vsearch(
            fasta_file=fasta_file,
            threshold=threshold,
            temp_dir=temp_dir,
            iddef=iddef,
            vsearch_bin=vsearch_bin,
            strand=strand,
            threads=threads,
            as_dict=True,
            debug=debug,
        )
    else:
        err = f"Invalid algo option: {algo}."
        err += " Valid choices are: 'vsearch', 'mmseqs', 'cdhit', or 'auto'."
        raise (ValueError, err)
    cluster_info = {}
    for cname, cdata in cluster_dict.items():
        cluster_info[cname] = {
            "centroid": seq_dict[cdata["centroid_id"]],
            "seqs": [seq_dict[seq_id] for seq_id in cdata["seq_ids"]],
        }
    if as_dict:
        return cluster_info
    return Clusters(cluster_info)


def cluster_vsearch(
    fasta_file: str,
    threshold: float = 0.975,
    temp_dir: str = "/tmp",
    iddef: int = 0,
    vsearch_bin: str = None,
    strand: str = "plus",
    threads: Optional[int] = None,
    as_dict: bool = False,
    debug: bool = False,
) -> Union[dict, Clusters]:
    """
    Clusters sequences using `VSEARCH <https://github.com/torognes/vsearch>`_.

    Parameters
    ----------
    fasta_file : string
        Path to a FASTA-formatted file. Required. If you'd like to run ``vsearch``
        using ``Sequence`` objects as input, use ``cluster(algo="vsearch")``.

    threshold : float, default=0.975
        Identity threshold for clustering. Must be between 0 and 1.

    temp_dir : str, default="/tmp"
        Path to a directory for temporary storage of clustering files.

    iddef : int, default=1
        Identity definition, as implemented by VSEARCH. Options are:
            0. CD-HIT definition: (matching columns) / (shortest sequence length).
            1. edit distance: (matching columns) / (alignment length).
            2. edit distance excluding terminal gaps (same as --id).
            3. Marine Biological Lab definition counting each gap opening (internal or
               terminal) as a single mismatch, whether or not the gap was extended: 1.0
               - [(mismatches + gap openings)/(longest sequence length)]
            4. BLAST definition, equivalent to --iddef 1 in a context of global pairwise
               alignment.

    vsearch_bin : str, optional
        Path to a VSEARCH executable. If not provided, the VSEARCH binary bundled
        with ``abutils`` will be used.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    strand : str, default="plus"
        Strand of the sequences to align. Options are ``"plus"`` and ``"both"``.

    as_dict : bool, default=False
        If ``True``, return clustering results as a ``dict`` rather than a ``Clusters``
        object. the ``dict`` is of the format:
            {"cluster1_name": {"centroid": cluster1_centroid,
                               "seqs": [seq1, seq2, seq3, ...]},
             "cluster2_name": {"centroid": cluster2_centroid,
                               "seqs": [seq4, seq5, seq6, ...]},
            }

    debug : bool, default=False
        If ``True``, prints standard output and standard error from ``vsearch``.
        Default is ``False``.


    Returns
    -------
    clusters : Path to the UC output file from ``vsearch`` or a ``dict`` of cluster info.

    """
    # output files
    centroid_file = tempfile.NamedTemporaryFile(
        dir=temp_dir, delete=False, prefix="centroids_"
    ).name
    uc_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False, prefix="uc_").name
    # get the vsearch binary
    if vsearch_bin is None:
        vsearch_bin = get_binary_path("vsearch")
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # system = platform.system().lower()
        # machine = platform.machine().lower().replace("x86_64", "amd64")
        # vsearch_bin = os.path.join(mod_dir, f"bin/vsearch_{system}_{machine}")
    # do clustering
    vsearch_cmd = f"{vsearch_bin} --cluster_fast {fasta_file}"
    vsearch_cmd += f" --centroids {centroid_file}"
    vsearch_cmd += " --clusterout_id"
    vsearch_cmd += f" --uc {uc_file}"
    vsearch_cmd += f" --id {threshold}"
    vsearch_cmd += f" --iddef {iddef}"
    vsearch_cmd += " --sizeout"
    vsearch_cmd += f" --strand {strand}"
    if threads is not None:
        vsearch_cmd += f" --threads {threads}"
    p = sp.Popen(vsearch_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        print("STDOUT:", stdout.decode("utf-8"))
        print("")
        print("STDERR:", stderr.decode("utf-8"))
    # process output
    if os.stat(uc_file).st_size == 0:
        err = f"WARNING: the VSEARCH output file ({uc_file}) is empty. "
        err += "Please verify that the input data is valid."
        print(err)
        return None
    if as_dict:
        cluster_info = {}
        with open(uc_file, "r") as f:
            for line in f:
                if not (ldata := line.strip().split()):
                    continue
                cluster_num = ldata[1]
                if cluster_num not in cluster_info:
                    cluster_info[cluster_num] = {"seq_ids": []}
                if ldata[0] == "S":
                    centroid = ldata[-2]
                    cluster_info[cluster_num]["centroid_id"] = centroid
                    cluster_info[cluster_num]["seq_ids"].append(centroid)
                elif ldata[0] == "H":
                    hit = ldata[-2]
                    cluster_info[cluster_num]["seq_ids"].append(hit)
        if not debug:
            os.remove(uc_file)
        return cluster_info
    else:
        return uc_file


def cluster_mmseqs(
    fasta_file: str,
    threshold: float = 0.975,
    cluster_mode: str = "2",
    cov_mode: str = "0",
    coverage: float = 0.8,
    alignment_mode: str = "3",
    seq_id_mode: str = "1",
    threads: Optional[int] = None,
    temp_dir: str = "/tmp",
    mmseqs_bin: str = None,
    as_dict: bool = False,
    debug: bool = False,
):
    """
    Clusters sequences using `MMseqs2 <https://github.com/soedinglab/MMseqs2>`_.

    Parameters
    ----------
    fasta_file : string
        Path to a FASTA-formatted file. Required. If you'd like to run ``mmseqs``
        using ``Sequence`` objects as input, use ``cluster(algo="mmseqs")``.

    threshold : float, default=0.975
        Identity threshold for clustering. Must be between 0 and 1.

    cluster_mode : str, default="2"
        Clustering mode. Options are ``"1"``, ``"2"``, or ``"3"``.
          - ``"1"``: greedy set cover
          - ``"2"``: connected component
          - ``"3"``: greedy incremental (CD-HIT-like)
        See the `MMseqs2 documentation <https://mmseqs.com/latest/userguide.html#clustering>`_
        for more information.

    cov_mode : str, default="0"
        Coverage mode. Options are ``"0"``, ``"1"``, ``"2"``, or ``"3"``.
            - ``"0"``: bidirectional
            - ``"1"``: target coverage
            - ``"2"``: query coverage
            - ``"3"``: target-in-query length coverage
        See the `MMseqs2 documentation <https://mmseqs.com/latest/userguide.html#clustering>`_
        for more information.

    coverage : float, default=0.8
        Coverage threshold for clustering. Must be between 0 and 1.

    alignment_mode : str, default="3"
        Alignment mode. Options are ``"0"``, ``"1"``, ``"2"``, ``"3"``, or ``"4"``.
            - ``"0"``: automatic
            - ``"1"``: only score and end_pos
            - ``"2"``: also start_pos and cov
            - ``"3"``: also seq.id
            - ``"4"``: only ungapped alignment
        See the `MMseqs2 documentation <https://mmseqs.com/latest/userguide.html#clustering>`_
        for more information.

    seq_id_mode : str, default="1"
        Sequence ID mode. Options are ``"0"``, ``"1"``, or ``"2"``.
            - ``"0"``: alignment length
            - ``"1"``: shorter sequence length
            - ``"2"``: longer sequence length

    threads : int, default=None
        Number of threads to use for clustering. If not provided, the number of threads
        will be determined by MMseqs2.

    temp_dir : str, default="/tmp"
        Path to a directory for temporary storage of clustering files.

    mmseqs_bin : str, optional
        Path to a MMseqs2 executable. If not provided, the MMseqs2 binary bundled
        with ``abutils`` will be used.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    as_dict : bool, default=False
        If ``True``, return clustering results as a ``dict`` rather than the TSV output file.
        object. the ``dict`` is of the format:
            {"cluster1_name": {"centroid_id": cluster1_centroid_id,
                               "seq_ids": [seq1_id, seq2_id, seq3_id, ...]},
             "cluster2_name": {"centroid_id": cluster2_centroid_id,
                               "seq_ids": [seq4_id, seq5_id, seq6_id, ...]},
            }

    debug : bool, default=False
        If ``True``, prints standard output and standard error from ``mmseqs``.
        Default is ``False``.


    Returns
    -------
    clusters : Path to the TSV output file from ``mmseqs`` or a ``dict`` of cluster info.

    """
    # output files
    db_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False, prefix="DB_").name
    clu_file = tempfile.NamedTemporaryFile(
        dir=temp_dir, delete=False, prefix="CLU_"
    ).name
    tsv_file = tempfile.NamedTemporaryFile(
        dir=temp_dir, delete=False, prefix="TSV_"
    ).name
    # get the mmseqs binary
    if mmseqs_bin is None:
        mmseqs_bin = get_binary_path("mmseqs")
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # system = platform.system().lower()
        # machine = platform.machine().lower().replace("x86_64", "amd64")
        # mmseqs_bin = os.path.join(mod_dir, f"bin/mmseqs_{system}_{machine}")
    # build the mmseqs DB
    db_cmd = f"{mmseqs_bin} createdb {fasta_file} {db_file}"
    p = sp.Popen(db_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        print("STDOUT:", stdout)
        print("")
        print("STDERR:", stderr)
    # do the clustering
    cluster_cmd = f"{mmseqs_bin} cluster"
    cluster_cmd += f" {db_file} {clu_file} {temp_dir}"
    cluster_cmd += f" --min-seq-id {threshold}"
    cluster_cmd += f" --cluster-mode {cluster_mode}"
    cluster_cmd += f" --cov-mode {cov_mode}"
    cluster_cmd += f" -c {coverage}"
    cluster_cmd += f" --alignment-mode {alignment_mode}"
    cluster_cmd += f" --seq-id-mode {seq_id_mode}"
    if threads is not None:
        cluster_cmd += f" --threads {threads}"
    p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        print("STDOUT:", stdout)
        print("")
        print("STDERR:", stderr)
    # generate TSV-formatted output
    tsv_cmd = f"{mmseqs_bin} createtsv"
    tsv_cmd += f" {db_file} {db_file} {clu_file} {tsv_file}"
    p = sp.Popen(tsv_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        print("STDOUT:", stdout)
        print("")
        print("STDERR:", stderr)
    # parse TSV output
    if as_dict:
        cluster_info = {}
        name_dict = {}
        with open(tsv_file) as f:
            for line in f:
                line = line.strip()
                if line.startswith("#"):
                    continue
                c, s = line.split()
                if c not in name_dict:
                    name = "".join(
                        random.choice(string.ascii_uppercase + string.digits)
                        for _ in range(8)
                    )
                    name_dict[c] = name
                name = name_dict[c]
                if name not in cluster_info:
                    cluster_info[name] = {"centroid_id": c, "seq_ids": []}
                cluster_info[name]["seq_ids"].append(s)
        if not debug:
            os.remove(tsv_file)
        return cluster_info
    else:
        return tsv_file


def cluster_cdhit(
    fasta_file: str,
    threshold: float = 0.975,
    temp_dir: str = "/tmp",
    cdhit_bin: str = None,
    threads: int = 0,
    as_dict: bool = False,
    debug: bool = False,
):
    """
    Clusters sequences using `MMseqs2`_.

    Parameters
    ----------
    fasta_file : string
        Path to a FASTA-formatted file. Required. If you'd like to run ``mmseqs``
        using ``Sequence`` objects as input, use ``cluster(algo="mmseqs")``.

    threshold : float, default=0.975
        Identity threshold for clustering. Must be between 0 and 1.

    temp_dir : str, default="/tmp"
        Path to a directory for temporary storage of clustering files.

    cdhit_bin : str, optional
        Path to a CD-HIT executable. If not provided, the CD-HIT binary bundled
        with ``abutils`` will be used.

    threads : int, default=0
        Number of threads to use for clustering. If not provided, all available CPU cores
        will be used.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    as_dict : bool, default=False
        If ``True``, return clustering results as a ``dict`` rather than the TSV output file.
        object. the ``dict`` is of the format:
            {"cluster1_name": {"centroid_id": cluster1_centroid_id,
                               "seq_ids": [seq1_id, seq2_id, seq3_id, ...]},
             "cluster2_name": {"centroid_id": cluster2_centroid_id,
                               "seq_ids": [seq4_id, seq5_id, seq6_id, ...]},
            }

    debug : bool, default=False
        If ``True``, prints standard output and standard error from ``mmseqs``.
        Default is ``False``.


    Returns
    -------
    clusters : Path to the `".clstr"` output file from CD-HIT or a ``dict`` of cluster info.


    .. _CD-HIT
        https://github.com/weizhongli/cdhit/wiki

    """
    # output files
    output_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False).name
    clusters_file = output_file + ".clstr"
    # get the CD-HIT binary
    if cdhit_bin is None:
        cdhit_bin = get_binary_path("cdhit")
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # system = platform.system().lower()
        # machine = platform.machine().lower().replace("x86_64", "amd64")
        # cdhit_bin = os.path.join(mod_dir, f"bin/cdhit_{system}_{machine}")
    # run CD-HIT
    # clustering options are as follows:
    #   - d: length at which to truncate sequence names in the output file,
    #        default 20 (0 to truncate at first space)
    #   - l: length of the throw_away_sequences, default 10
    #   - T: number of threads, default 1 (0 uses all available threads)
    #   - M: max memory (in MB), default 800 (0 uses all available memory)
    wordsize = _get_cdhit_wordsize(threshold)
    cluster_cmd = f"{cdhit_bin} -i {fasta_file} -o {output_file}"
    cluster_cmd += f" -c {threshold} -n {wordsize} -T {threads}"
    cluster_cmd += " -d 0 -l 4 -M 0"
    p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        print("STDOUT:", stdout)
        print("")
        print("STDERR:", stderr)
    if os.stat(clusters_file).st_size == 0:
        err = f"WARNING: the CD-HIT output file ({clusters_file}) is empty. "
        err += "Please verify that the input data is valid."
        print(err)
        return None
    # parse CD-HIT output
    if as_dict:
        cluster_info = {}
        with open(clusters_file) as f:
            raw_clusters = f.read().split("\n>")
        for raw_cluster in raw_clusters:
            if not raw_cluster.strip():
                continue
            cluster_lines = raw_cluster.split("\n")
            centroid_id = None
            cluster_name = cluster_lines[0]
            seq_ids = []
            for cluster_line in cluster_lines[1:]:
                if not cluster_line.strip():
                    continue
                seq_id = cluster_line.split(">")[1].split("...")[0]
                seq_ids.append(seq_id)
                if "*" in cluster_line:
                    centroid_id = seq_id
            if centroid_id is None:
                centroid_id = seq_ids[0]
            cluster_info[cluster_name] = {
                "centroid_id": centroid_id,
                "seq_ids": seq_ids,
            }
        if not debug:
            os.remove(output_file)
            os.remove(clusters_file)
        return cluster_info
    else:
        return clusters_file


def _get_cdhit_wordsize(threshold):
    if threshold > 0.7:
        return 5
    elif threshold > 0.6:
        return 4
    elif threshold > 0.5:
        return 3
    return 2


def linclust(
    seqs: Union[str, Iterable[Sequence]],
    threshold: float = 0.975,
    cov_mode: Union[str, int] = 0,
    out_file=None,
    temp_dir=None,
    make_db=True,
    linclust_kwargs: str = None,
    quiet=False,
    threads=0,
    debug=False,
):
    """
    Perform clustering of sequences using ``mmseqs easy-linclust``.
    """
    pass


def nested_linclust(
    sequences: Union[str, Iterable[Union[Sequence, Pair]]],
    thresholds: Iterable[float],
    project_directory: str,
    linclust_kwargs: Optional[Union[str, dict]] = None,
    parquet: bool = False,
    verbose: bool = True,
    debug: bool = False,
) -> None:
    """
    Perform nested clustering of sequences using ``mmseqs easy-linclust``.

    Parameters
    ----------
    sequences : Union[str, Iterable[Union[Sequence, Pair]]
        If a ``str`` is provided, it is assumed to be the path to a AIRR-formatted file (or a
        directory of AIRR-formatted files) containing the sequences to be clustered. This file
        should be tab-delimited unless `parquet` is ``True``, in which case it should be
        a ``parquet`` file. If an iterable is provided, it can contain ``Sequence`` objects,
        ``Pair`` objects, or a mixture of both. If pairs are provided, the heavy/light chains
        will be concatenated prior to clustering.

    thresholds : Iterable[float]
        An iterable of thresholds for clustering. Each should be between 0 and 1.
        Thresholds will be sorted in descending order. The first threshold will be used
        for the initial clustering, the results will be clustered at the second threshold,
        and so on.

    project_directory : str
        The path to the project directory. The necessary directory structure will be
        created automatically based on the provided thresholds.

    linclust_kwargs : Optional[str], optional
        Keyword arguments to pass to ``mmseqs easy-linclust``. Should be a string of the form
        ``"--arg1 value1 --flag1, --arg2 value2..."``, or can be a dictionary mapping
        threshold values to their corresponding arguments. If a ``dict`` is passed and some
        thresholds are missing, the missing thresholds will not pass any additional kwargs
        to ``mmseqs easy-linclust``.
    """

    # linclust kwargs
    if isinstance(linclust_kwargs, str):
        linclust_kwargs = {t: linclust_kwargs for t in thresholds}


# def cluster(seqs, threshold=0.975, out_file=None, temp_dir=None, make_db=True, method='cdhit',
#             quiet=False, threads=0, return_just_seq_ids=False, max_memory=800, retries=5, debug=False):
#     '''
#     Perform sequence clustering with CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that ``abutils.utils.sequence.Sequence``
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is ``0.975``.

#         out_file (str): Path to the clustering output file. Default is to use
#             ``tempfile.NamedTemporaryFile`` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, ``'/tmp'`` is used.

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs)
#             Default is ``True``.

#         method (str): Doesn't do anything at this point, stubbed to allow expansion of sequencing tools
#             beyond CD-HIT (UCLUST, etc).

#         quiet (bool): If ``True``, surpresses printing of output/progress info. Default is ``False``.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is ``0``,
#             which results in all available cores being used.

#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the ``-M`` runtime option. Default is ``800``.

#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty.
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are
#             not empty. Default is 5.

#         debug (bool): If ``True``, print standard output and standard error from CD-HIT. Default is ``False``.

#     Returns:

#         list: A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     if make_db:
#         ofile, cfile, seq_db, db_path = cdhit(seqs, out_file=out_file, temp_dir=temp_dir,
#                                               threshold=threshold, make_db=True, quiet=quiet,
#                                               threads=threads, max_memory=max_memory, debug=debug)
#         return parse_clusters(ofile, cfile, seq_db=seq_db, db_path=db_path, return_just_seq_ids=return_just_seq_ids)
#     else:
#         if isinstance(seqs, (list, tuple)):
#             seqs = [Sequence(s) for s in seqs]
#             seq_dict = {s.id: s for s in seqs}
#         elif os.path.exists(seqs):
#             seq_dict = None
#         ofile, cfile, = cdhit(seqs, out_file=out_file, temp_dir=temp_dir, threads=threads,
#                               threshold=threshold, make_db=False, quiet=quiet,
#                               max_memory=max_memory, debug=debug)
#         return parse_clusters(ofile, cfile, seq_dict=seq_dict, return_just_seq_ids=return_just_seq_ids)


# def cdhit(seqs, out_file=None, temp_dir=None, threshold=0.975, make_db=True,
#           quiet=False, threads=0, max_memory=800, retries=5, debug=False):
#     '''
#     Run CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that `abutils.utils.sequence.Sequence()`
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a file path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is `0.975`.

#         out_file (str): Path to the clustering output file. Default is to use
#             `tempfile.NamedTemporaryFile` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, the default
#             temporary directory is used (typically ``/tmp`` on MacOS and Linux).

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs).
#             Default is `True`.

#         quiet (bool): If `True`, surpresses printing of output/progress info. Default is `False`.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is `0`,
#             which results in all available cores being used.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the `-M` runtime option. Default is `800`.

#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty.
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are
#             not empty. Default is 5.

#         debug (bool): If `True`, print standard output and standard error from CD-HIT. Default is `False`.

#     Returns:

#         If `make_db` is `True`, returns the CD-HIT output file path, the CD-HIT cluster file path,
#             a `sqlite3` database connection object, and the database path. If `make_db` is `False`, only the
#             CD-HIT output file path and CD-HIT cluster file path are returned.
#     '''
#     start_time = time.time()
#     delete_input = False if debug else True
#     if not quiet:
#         print('CD-HIT: clustering {} seqeunces'.format(len(seqs)))
#     if out_file is None:
#         out_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#         out_file.close()
#         ofile = out_file.name
#     else:
#         ofile = os.path.expanduser(out_file)
#     cfile = ofile + '.clstr'
#     with open(ofile, 'w') as f: f.write('')
#     with open(cfile, 'w') as f: f.write('')
#     if isinstance(seqs, (list, tuple)):
#         seqs = [Sequence(s) for s in seqs]
#         ifile = _make_cdhit_input(seqs, temp_dir)
#     elif os.path.exists(seqs):
#         ifile = seqs
#         delete_input = False
#     word_size = _get_cdhit_wordsize(threshold)
#     cdhit_cmd = 'cdhit -i {} -o {} -c {} -n {} -d 0 -T {} -M {}'.format(ifile,
#                                                                         ofile,
#                                                                         threshold,
#                                                                         word_size,
#                                                                         threads,
#                                                                         max_memory)
#     while not all([os.path.getsize(cfile), os.path.getsize(cfile)]):
#         cluster = sp.Popen(cdhit_cmd,
#                         shell=True,
#                         stdout=sp.PIPE,
#                         stderr=sp.PIPE)
#         stdout, stderr = cluster.communicate()
#         if not retries:
#             break
#         retries -= 1
#     end_time = time.time()
#     if debug:
#         print(stdout)
#         print(stderr)
#     if delete_input:
#         os.unlink(ifile)
#     if not quiet:
#         print('CD-HIT: clustering took {:.2f} seconds'.format(end_time - start_time))
#     if make_db:
#         if not quiet:
#             print('CD-HIT: building a SQLite3 database')
#         seq_db, db_path = _build_seq_db(seqs, direc=temp_dir)
#         return ofile, cfile, seq_db, db_path
#     return ofile, cfile


# def parse_clusters(out_file, clust_file, seq_db=None, db_path=None, seq_dict=None, return_just_seq_ids=False):
#     '''
#     Parses CD-HIT output.

#     Args:

#         out_file (str): Path to the CD-HIT output file. Required.

#         clust_file (str): Path to the CD-HIT cluster file. Required.

#         seq_db (sqlite.Connection): SQLite3 `Connection` object. Default is ``None``. If not provided and
#             ``return_just_seq_ids`` is False, the returned ``Cluster`` objects will not contain any sequence
#             information beyond the sequence ID.

#         db_path (str): Path to a SQLite3 database file. Default is ``None``. Must be provided if
#             ``seq_db`` is also provided.

#         seq_dict (dict): A ``dict`` mapping sequence IDs to ``abutils.core.sequence.Sequence`` objects. Default
#             is ``None``. Typically used when a relatively small number of sequences are being clustered and
#             creating a ``sqlite3`` database would be overkill.

#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#     Returns:

#         A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     raw_clusters = [c.split('\n') for c in open(clust_file, 'r').read().split('\n>')]
#     if return_just_seq_ids:
#         ids = []
#         for rc in raw_clusters:
#             _ids = []
#             for c in rc[1:]:
#                 if c:
#                     _ids.append(c.split()[2][1:-3])
#             ids.append(_ids)
#         os.unlink(out_file)
#         os.unlink(clust_file)
#         return ids
#     os.unlink(out_file)
#     os.unlink(clust_file)
#     clusters = [Cluster(rc, seq_db, db_path, seq_dict) for rc in raw_clusters]
#     return CDHITResult(clusters, seq_db=seq_db, db_path=db_path, seq_dict=seq_dict)


# def _make_cdhit_input(seqs, temp_dir):
#     ifile = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#     ifile.close()
#     fastas = [s.fasta for s in seqs]
#     with open(ifile.name, 'w') as f:
#         f.write('\n'.join(fastas))
#     return ifile.name


# def _get_cdhit_wordsize(threshold):
#     if threshold > 0.7:
#         return 5
#     elif threshold > 0.6:
#         return 4
#     elif threshold > 0.5:
#         return 3
#     return 2


# def _build_seq_db(seqs, direc=None):
#     # '''
#     # Builds a SQLite3 database of sequences.

#     # Inputs are a list of Sequence objects and an optional directory to store the database.
#     # If ::direc:: is not provided, '/tmp' will be used.

#     # Returns a SQLite3 connection object and the database path.
#     # '''
#     direc = direc if direc is not None else '/tmp'
#     db_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
#     db_path = os.path.join(direc, db_name)
#     conn = sqlite3.connect(db_path)
#     c = conn.cursor()
#     create_cmd = '''CREATE TABLE seqs (id text, sequence text)'''
#     insert_cmd = 'INSERT INTO seqs VALUES (?,?)'
#     c.execute('DROP TABLE IF EXISTS seqs')
#     c.execute(create_cmd)
#     c.executemany(insert_cmd, [(str(s.id), str(s.sequence)) for s in seqs])
#     c.execute('CREATE INDEX seq_index ON seqs (id)')
#     return c, db_path



#!/usr/bin/env python
# filename: mongodb.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import platform
import os
import subprocess as sp
import sys

from pymongo import MongoClient

from . import log

if sys.version_info[0] > 2:
    STR_TYPES = [
        str,
    ]
else:
    STR_TYPES = [str, unicode]


def get_connection(ip="localhost", port=27017, user=None, password=None):
    """
    Returns a pymongo ``MongoClient`` object.

    .. note:

        Both ``user`` and ``password`` are required when connecting to a MongoDB
        database that has authentication enabled.

    Arguments:

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (int): Port of the MongoDB server. Default is ``27017``.

        user (str): Username, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.

        password (str): Password, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.
    """
    if platform.system().lower() == "darwin":
        connect = False
    else:
        connect = True
    if user and password:
        import urllib

        pwd = urllib.quote_plus(password)
        uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
        return MongoClient(uri, connect=connect)
    return MongoClient(ip, port, connect=connect)


def list_dbs(ip="localhost", port=27017, user=None, connect=True, hint=None):
    """
    Returns a list of Databases.

    .. note:

        Both ``user`` and ``password`` are required when connecting to a MongoDB
        database that has authentication enabled.

    Arguments:

        db (str): Name of the MongoDB database. Required.

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (int): Port of the MongoDB server. Default is ``27017``.

        user (str): Username, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.

        password (str): Password, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.
            
        hint (str): substring found in database name, if used only list of 
            databases that contain the substring will returned
    """

    # Darwin says "you shall not connect!"
    if platform.system().lower() == "darwin":
        connect = False
    else:
        connect = True

    # check for user and pass, if so: make uri, import urllib, return dbs
    if user and password:
        import urllib

        pwd = urllib.quote_plus(password)
        uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
        # check to see if hint was passed in and return list of dbs with hint
        if hint and type(hint) == str:
            ls = MongoClient(uri, connect=connect).list_database_names()
            return [d for d in ls if hint.upper() in d.upper()]
        # otherwise return a list of all dbs
        else:
            return MongoClient(uri, connect=connect).list_database_names()

    # otherwise no user or pass required
    else:
        # check to see if hint was passed in and return list of dbs with hint
        if hint and type(hint) == str:
            ls = MongoClient(ip, port, connect=connect).list_database_names()
            return [d for d in ls if hint.upper() in d.upper()]
        # otherwise return a list of all dbs
        else:
            return MongoClient(ip, port, connect=connect).list_database_names()


def get_db(db, ip="localhost", port=27017, user=None, password=None):
    """
    Returns a pymongo ``Database`` object.

    .. note:

        Both ``user`` and ``password`` are required when connecting to a MongoDB
        database that has authentication enabled.

    Arguments:

        db (str): Name of the MongoDB database. Required.

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (int): Port of the MongoDB server. Default is ``27017``.

        user (str): Username, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.

        password (str): Password, if authentication is enabled on the MongoDB database.
            Default is ``None``, which results in requesting the connection
            without authentication.
    """
    if platform.system().lower() == "darwin":
        connect = False
    else:
        connect = True
    if user and password:
        import urllib

        pwd = urllib.quote_plus(password)
        uri = "mongodb://{}:{}@{}:{}".format(user, pwd, ip, port)
        conn = MongoClient(uri, connect=connect)
    else:
        conn = MongoClient(ip, port, connect=connect)
    return conn[db]


def get_collections(db, collection=None, prefix=None, suffix=None):
    """
    Returns a sorted list of collection names found in ``db``.

    Arguments:

        db (Database): A pymongo Database object. Can be obtained
            with ``get_db``.

        collection (str): Name of a collection. If the collection is
            present in the MongoDB database, a single-element list will
            be returned with the collecion name. If not, an empty list
            will be returned. This option is primarly included to allow
            for quick checking to see if a collection name is present.
            Default is None, which results in this option being ignored.

        prefix (str): If supplied, only collections that begin with
            ``prefix`` will be returned.

        suffix (str): If supplied, only collections that end with
            ``suffix`` will be returned.

    Returns:

        list: A sorted list of collection names.
    """
    if collection is not None:
        return [
            collection,
        ]
    collections = db.collection_names(include_system_collections=False)
    if prefix is not None:
        collections = [c for c in collections if c.startswith(prefix)]
    if suffix is not None:
        collections = [c for c in collections if c.endswith(suffix)]
    return sorted(collections)


def rename_collection(db, collection, new_name):
    """
    Renames a MongoDB collection.

    Arguments:

        db (Database): A pymongo ``Database`` object. Can be obtained
            with ``get_db``.

        collection (str): Name of the collection to be renamed.

        new_name (str, func): ``new_name`` can be one of two things::

            1. The new collection name, as a string.
            2. A function which, when passed the current collection name,
                returns the new collection name. If the function
                returns an empty string, the collection will not be
                renamed.
    """
    if hasattr(new_name, "__call__"):
        _new = new_name(collection)
        if _new == "":
            return
    else:
        _new = new_name
    c = db[collection]
    c.rename(_new)


def update(field, value, db, collection, match=None):
    """
    Updates MongoDB documents.

    Sets ``field`` equal to ``value`` for all documents that
    meet ``match`` criteria.

    Arguments:

        field (str): Field to update.

        value (str): Update value.

        db (Database): A pymongo ``Database`` object.

        collection (str): Collection name.

        match (dict): A dictionary containing the match criteria, for example::

            {'seq_id': {'$in': ['a', 'b', 'c']}, 'cdr3_len': {'$gte': 18}}
    """
    c = db[collection]
    match = match if match is not None else {}
    # check MongoDB version to use appropriate update command
    if db.client.server_info()["version"].startswith("2"):
        c.update(match, {"$set": {field: value}}, multi=True)
    else:
        c.update_many(match, {"$set": {field: value}})


def unset(db, collection, field, match=None):
    """
    Removes ``field`` from all records in ``collection`` that meet
    ``match`` criteria.

    Arguments:

        field (str): Field to be removed.

        db (Database): A pymongo Database object.

        collection (str): Collection name.

        match (dict): A dictionary containing the match criteria, for example::

            {'seq_id': {'$in': ['a', 'b', 'c']}, 'cdr3_len': {'$gte': 18}}
    """
    c = db[collection]
    match = match if match is not None else {}
    # check MongoDB version to use appropriate update command
    if db.client.server_info()["version"].startswith("2"):
        c.update(match, {"$unset": {field: ""}}, multi=True)
    else:
        c.update_many(match, {"$unset": {field: ""}})


def mongoimport(
    json,
    database,
    ip="localhost",
    port=27017,
    user=None,
    password=None,
    delim="_",
    delim1=None,
    delim2=None,
    delim_occurance=1,
    delim1_occurance=1,
    delim2_occurance=1,
):
    """
    Performs mongoimport on one or more json files.

    Args:

        json: Can be one of several things:

            - path to a single JSON file
            - an iterable (list or tuple) of one or more JSON file paths
            - path to a directory containing one or more JSON files

        database (str): Name of the database into which the JSON files
            will be imported

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (int): Port of the MongoDB database. Default is ``27017``.

        user (str): Username for the MongoDB database, if authentication is enabled.
            Default is ``None``, which results in attempting connection without
            authentication.

        password (str): Password for the MongoDB database, if authentication is enabled.
            Default is ``None``, which results in attempting connection without
            authentication.

        delim (str): Delimiter, when generating collection names using a single delimiter.
            Default is ``_``

        delim_occurance (int): Occurance at which to split filename when using a
            single delimiter. Default is ``1``

        delim1 (str): Left delimiter when splitting with two delimiters. Default is None.

        delim1_occurance (int): Occurance of ``delim1`` at which to split filename.
            Default is ``1``

        delim2 (str): Right delimiter when splitting with two delimiters. Default is None.

        delim2_occurance (int): Occurance of ``delim2`` at which to split filename.
            Default is ``1``
    """
    logger = log.get_logger("mongodb")
    _print_mongoimport_info(logger)
    if type(json) in (list, tuple):
        pass
    elif os.path.isdir(json):
        from abtools.utils.pipeline import list_files

        json = list_files(json)
    else:
        json = [
            json,
        ]
    jsons = sorted([os.path.expanduser(j) for j in json if j.endswith(".json")])
    collections = _get_import_collections(
        jsons,
        delim,
        delim_occurance,
        delim1,
        delim1_occurance,
        delim2,
        delim2_occurance,
    )
    logger.info("Found {} files to import".format(len(jsons)))
    logger.info("")
    for i, (json_file, collection) in enumerate(zip(jsons, collections)):
        logger.info(
            "[ {} ] {} --> {}".format(i + 1, os.path.basename(json_file), collection)
        )
        # logger.info("Performing mongoimport on {}.".format(os.path.basename(json_file)))
        # logger.info("Importing the file into collection {}.".format(collection))
        if all([user, password]):
            host = "--host {} --port {} -username {} -password {}".format(
                ip, port, user, password
            )
        else:
            host = "--host {} --port {}".format(ip, port)
        mongo_cmd = "mongoimport {} --db {} --collection {} --file {}".format(
            host, database, collection, json_file
        )
        mongo = sp.Popen(mongo_cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = mongo.communicate()


def index(db, collection, fields, directions=None, desc=False, background=False):
    """
    Builds a simple (single field) or complex (multiple fields) index
    on a single collection in a MongoDB database.

    Args:

        db (Database): A pymongo Database object.

        collection (str): Collection name.

        fields: Can be one of two things:

            - the name of a single field, as a string
            - an iterable (list/tuple) of one or more field names

        desc (bool): If ``True``, all indexes will be created in descending order.
            Default is ``False``.

        directions (list): For complex indexes for which you'd like to have
            different indexing directions (ascending for some fields, descending
            for others), you can pass a list of pymongo direction objects (
            pymongo.ASCENDING and pymongo.DESCENDING), in the same order as the
            list of fields to be indexed. Must be the same length as the list
            of index fields. Default is ``None``.

        background (bool): If ``True``, the indexing operation will be processed
            in the background. When performing background indexes, the MongoDB
            database will not be locked.
    """
    import pymongo

    if type(fields) in STR_TYPES:
        fields = [
            fields,
        ]
    if directions is None:
        _dir = pymongo.DESCENDING if desc else pymongo.ASCENDING
        directions = [_dir] * len(fields)
    field_tuples = list(zip(fields, directions))
    coll = db[collection]
    coll.create_index(field_tuples, background=background)


def remove_padding(db, collection, field="padding"):
    """
    Removes a padding field.

    Args:

        db (Database): A pymongo Database object.

        collection (str): Collection name

        field (str): Name of the padding field. Default is ``padding``
    """
    unset(db, collection, field=field)
    # c = db[collection]
    # c.update({}, {'$unset': {field: ''}}, multi=True)


def _get_import_collections(
    jsons, delim, delim_occurance, delim1, delim1_occurance, delim2, delim2_occurance
):
    jnames = [os.path.basename(j) for j in jsons]
    if not all([delim1, delim2]):
        collections = [delim.join(j.split(delim)[:delim_occurance]) for j in jnames]
    else:
        pre_colls = [delim1.join(j.split(delim1)[delim1_occurance:]) for j in jnames]
        collections = [
            delim2.join(j.split(delim2)[:delim2_occurance]) for j in pre_colls
        ]
    return collections


def _print_mongoimport_info(logger):
    logger.info("")
    logger.info("")
    logger.info("")
    logger.info("-" * 25)
    logger.info("MONGOIMPORT")
    logger.info("-" * 25)
    logger.info("")


def _print_remove_padding():
    logger = log.get_logger("mongodb")
    logger.info("Removing MongoDB padding...")




#!/usr/bin/env python
# filename: data.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import sys
from typing import Callable, Iterable, Optional, Union

import numpy as np
import pandas as pd

from scipy import stats

from ...core.sequence import Sequence


class PlotData:
    """
    Provides methods for processing input data.
    """

    def __init__(
        self,
        x: Union[Iterable, str, None] = None,
        y: Union[Iterable, str, None] = None,
        data: Union[pd.DataFrame, Iterable, dict, None] = None,
        sequences: Optional[Iterable[Sequence]] = None,
        hue: Union[Iterable, str, None] = None,
        column_labels: Optional[Iterable] = None,
        row_labels: Optional[Iterable] = None,
    ):
        """
        Initializes a PlotData object.

        Parameters
        ----------
        data : Union[pd.DataFrame, Iterable, None], optional
            Input data. Can be a ``DataFrame``, a ``dict``, or an iterable. If both `data` and `sequences`
            are provided, only `sequences` is used. If neither `data` nor `sequences` are provided, the
            x, y, and hue data must be provided as iterables to `x`, `y`, and `hue`, respectively.

        sequences : Optional[Iterable[Sequence]], optional
            Input data. Must be an iterable of ``abutils.Sequence`` objects. If both `data` and `sequences`
            are provided, only `sequences` is used. If neither `data` nor `sequences` are provided, the
            x, y, and hue data must be provided as iterables to `x`, `y`, and `hue`, respectively.

        x : Union[Iterable, str, None], optional
            The x-axis data. Can be an iterable, a string, or ``None``. If an iterable, the length should
            match the length of all other provided inputs (`data`, `y`, and `hue`). If a string, it should
            be the name of a column in `data`.

        y : Union[Iterable, str, None], optional
            The y-axis data. Can be an iterable, a string, or ``None``. If an iterable, the length should
            match the length of all other provided inputs (`data`, `x`, and `hue`). If a string, it should
            be the name of a column in `data`.

        hue : Union[Iterable, str, None], optional
            The hue data. Can be an iterable, a string, or ``None``. If an iterable, the length should
            match the length of all other provided inputs (`data`, `x`, and `y`). If a string, it should
            be the name of a column in `data`.

        column_labels : Optional[Iterable], optional
            Labels for the columns of `data`. If provided, the length should match the number of columns
            in `data`.

        row_labels : Optional[Iterable], optional
            Labels for the rows of `data`. If provided, the length should match the number of rows in
            `data`.

        """
        self.raw_data = data
        self.raw_sequences = sequences
        self.raw_x = x
        self.raw_y = y
        self.raw_hue = hue
        self.raw_column_labels = column_labels
        self.raw_row_labels = row_labels

        self.x_name = None
        self.y_name = None
        self.hue_name = None

        self.raw_df = self.process_input(
            x=x,
            y=y,
            data=data,
            sequences=sequences,
            hue=hue,
            column_labels=column_labels,
            row_labels=row_labels,
        )
        self.df = self.raw_df.copy()

    @property
    def transform_funcs(self) -> dict:
        funcs = {
            "log": np.log,
            "log2": np.log2,
            "log10": np.log10,
            "log1p": np.log1p,
            "exp": np.exp,
            "linear": lambda x: x,
        }
        return funcs

    @property
    def agg_funcs(self) -> dict:
        funcs = {
            "mean": "mean",
            "average": "mean",
            "median": "median",
            "mode": pd.Series.mode,
            "sum": "sum",
            "count": "count",
            "size": "size",
            "nunique": "nunique",
            "max": "max",
            "min": "min",
            "idxmax": "idxmax",
            "idxmin": "idxmin",
            "first": "first",
            "last": "last",
            "cov": "cov",
            "sem": "sem",
            "std": "std",
            "var": "var",
        }
        return funcs

    def process_input(
        self,
        x: Union[str, Iterable, None] = None,
        y: Union[str, Iterable, None] = None,
        data: Union[pd.DataFrame, Iterable, dict, None] = None,
        sequences: Optional[Iterable[Sequence]] = None,
        hue: Union[str, Iterable, None] = None,
        column_labels: Optional[Iterable] = None,
        row_labels: Optional[Iterable] = None,
    ) -> pd.DataFrame:
        """
        Reads input datasets and returns a ``DataFrame`` containing x, y, and hue data.
        """
        # input can be provided as iterables passed to x, y and/or hue
        # the column names for x, y and hue are set to 'x', 'y', and 'hue', respectively
        if all([data is None, sequences is None]):
            _data = {}
            _data["x"] = x
            self.x_name = "x"
            _data["y"] = y
            self.y_name = "y"
            if hue is not None:
                _data["hue"] = hue
                self.hue_name = "hue"
            df = pd.DataFrame(_data)

        # input can also be passed to data as a DataFrame or iterable
        # or to sequences as an iterable of abutils.Sequence objects
        else:
            if sequences is not None:
                df = [s.annotations for s in sequences]
            elif isinstance(data, pd.DataFrame):
                df = data.copy()
            else:
                df = pd.DataFrame(data)
            # if any of the other inputs are iterables, add them to the DataFrame
            # also, set the x, y, and hue names
            if x is not None:
                if not isinstance(x, str) and len(x) == df.shape[0]:
                    df["x"] = x
                    self.x_name = "x"
                elif isinstance(x, str):
                    self.x_name = x
            if y is not None:
                if not isinstance(y, str) and len(y) == df.shape[0]:
                    df["y"] = y
                    self.y_name = "y"
                elif isinstance(y, str):
                    self.y_name = y
            if hue is not None:
                if not isinstance(hue, str) and len(hue) == df.shape[0]:
                    df["hue"] = hue
                    self.hue_name = "hue"
                elif isinstance(hue, str):
                    self.hue_name = hue

        # drop any columns that aren't x, y, or hue
        df = df[[self.x_name, self.y_name, self.hue_name]]

        # set column/row labels
        if column_labels is not None and len(column_labels) == df.shape[1]:
            df.columns = column_labels
        if row_labels is not None and len(row_labels) == df.shape[0]:
            df.index = row_labels
        return df

    def dropna(
        self,
        axis: Union[int, str] = "rows",
        how: str = "any",
        thresh: Optional[int] = None,
        subset: Optional[Union[str, Iterable]] = None,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Drop missing or NA values from the data. Default settings will drop all rows with any
        missing values. See the `pandas.DataFrame.dropna`_ documentation for more information.

        .. _pandas.DataFrame.dropna:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.dropna.html

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        df.dropna(axis=axis, how=how, thresh=thresh, subset=subset, inplace=True)
        if inplace:
            self.df = df
        else:
            return df

    def fillna(
        self,
        value: Optional[Union[int, float, str]] = None,
        axis: Union[int, str] = "columns",
        subset: Optional[Union[str, Iterable]] = None,
        method: Optional[str] = None,
        limit: Optional[int] = None,
        downcast: Optional[dict] = None,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Fill missing or NA values in the data. The default settings will fill all missing values
        with ``0``. See the `pandas.DataFrame.fillna`_ documentation for more information.

        Parameters
        ----------
        value : Optional[Union[int, float, str]]
            Value to use to fill holes. If ``None``, the ``method`` parameter must be specified.

        axis : Union[int, str]
            Only used in combination with `subset`. If ``0``, ``"column"``, or ``"columns"``,
            columns in `subset` will be filled. If ``1``, ``"row"``, or ``"rows"``, rows in
            `subset` will be filled. Default is ``"columns"``, so columns in `subset` will be
            filled.

        subset : Optional[Union[str, Iterable]]
            Subset of columns or rows to fill. If ``None``, all columns or rows will be filled.

        method : Optional[str]
            Method to use to fill holes. See the `pandas.DataFrame.fillna`_ documentation for
            available methods.

        limit : Optional[int]
            Maximum number of consecutive NaN values to forward/backward fill.

        downcast : Optional[dict]
            A dict of item->dtype of what to downcast if possible, or the string "infer" which
            will try to downcast to an appropriate equal type (e.g. float64 to int64 if possible).
            See the `pandas.DataFrame.fillna`_ documentation for more information.

        use_raw : bool
            If ``True``, fill missing values in the raw data. If ``False``, fill missing values in
            the processed data. Default is ``False``

        inplace : bool
            If ``True``, fill missing values in place. If ``False``, return a copy of the DataFrame
            with missing values filled. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, returns a copy of the DataFrame with missing values filled.
            Otherwise, returns ``None``.


        .. _pandas.DataFrame.fillna:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.fillna.html

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        if method is None:
            value = 0 if value is None else value

        # fill columns
        if axis in [0, "column", "columns"]:
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for s in subset:
                df[s].fillna(
                    value=value,
                    method=method,
                    limit=limit,
                    downcast=downcast,
                    inplace=True,
                )

        # fill rows
        if axis in [1, "row", "rows"]:
            df = df.T
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for s in subset:
                df[s].fillna(
                    value=value,
                    method=method,
                    limit=limit,
                    downcast=downcast,
                    inplace=True,
                )
            df = df.T

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def norm(
        self,
        subset: Union[Iterable, str, None] = None,
        axis: Union[int, str] = "columns",
        as_percent: bool = False,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Normalize the data. By default, each column is normalized to sum to 1.

        Parameters
        ----------
        subset : Union[Iterable, str, None]
            The columns or rows to normalize. If ``None``, all columns or rows will be normalized.
            Default is ``None``.

        axis : Union[int, str]
            The axis to normalize. If ``0``, ``"column"``, or ``"columns"``, each column will
            be  normalized. If ``1``, ``"row"``, or ``"rows"``, each row will be normalized.
            Default is ``"columns"``.

        as_percent : bool
            If ``True``, the data will be multiplied by ``100`` after normalization. Default is
            ``False``.

        use_raw : bool
            If ``True``, the raw data will be normalized. Otherwise, the processed data will be
            normalized. Default is ``False``.

        inplace : bool
            If ``True``, the normalized data will replace ``self.df``. Otherwise, the normalized
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the normalized ``DataFrame`` will be returned.

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        if isinstance(subset, str):
            subset = [subset]
        multiplier = 100 if as_percent else 1

        # normalize each column separately
        if axis in [0, "columns", "column"]:
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = df[c] / df[c].sum() * multiplier

        # normalize each row separately
        if axis in [1, "rows", "row"]:
            df = df.T
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = df[c] = df[c] / df[c].sum() * multiplier
            df = df.T

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def scale(
        self,
        subset: Union[Iterable, str, None] = None,
        axis: Union[int, str, None] = None,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Scale the data to be between 0 and 1. By default, the entire dataset is scaled
        according to the global minimum and maximum values in the dataset. Alternatively, each
        column or row can be scaled separately.

        Parameters
        ----------
        subset : Union[Iterable, str, None]
            The columns or rows to scale. Can be either a single row/column name, or an iterable
            of row/column names. If ``None``, all columns or rows will be scaled. Default is
            ``None``.

        axis : Union[int, str, None]
            The axis to scale. If ``0``, ``"column"``, or ``"columns"``, each column will be
            scaled separately. If ``1``, ``"row"``, or ``"rows"``, each row will be scaled separately.
            If ``None``, the entire dataset will be scaled. Default is ``None``.

        use_raw : bool
            If ``True``, the raw data will be scaled. Otherwise, the processed data will be scaled.
            Default is ``False``.

        inplace : bool
            If ``True``, the scaled data will replace ``self.df``. Otherwise, the scaled
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the scaled ``DataFrame`` will be returned.

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        if isinstance(subset, str):
            subset = [subset]

        # scale across the entire entire dataframe
        if axis is None:
            df = (df - df.min().min()) / (df.max().max() - df.min().min())

        # scale each column separately
        if axis in [0, "columns", "column"]:
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = (df[c] - df[c].min()) / (df[c].max() - df[c].min())

        # scale each row separately
        if axis in [1, "rows", "row"]:
            df = df.T
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = (df[c] - df[c].min()) / (df[c].max() - df[c].min())
            df = df.T

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def transform(
        self,
        func: Union[Callable, str, None] = None,
        subset: Union[Iterable, str, None] = None,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Transform the data (log, expoonential, etc). By default, the function is applied
        to all columns the dataset. Alternatively, a subset of columns can be transformed.

        Parameters
        ----------
        func : Union[Callable, str, None]
            The function to apply to the data. If ``None``, the data will not be transformed.
            If a string, the function must be one of the following:
                * ``"log"``: natural log
                * ``"log10"``: log base 10
                * ``"log2"``: log base 2
                * ``"log1p"``: natural log of (value + 1)
                * ``"exp"``: exponential
                * ``"linear"``: no transformation

        subset : Union[Iterable, str, None]
            The columns to transform. Can be either a single column name, or an iterable
            of column names. If ``None``, all columns will be transformed. Default is
            ``None``.

        use_raw : bool
            If ``True``, the raw data will be transformed. Otherwise, the processed data will be
            transformed. Default is ``False``.

        inplace : bool
            If ``True``, the transformed data will replace ``self.df``. Otherwise, the transformed
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the transformed ``DataFrame`` will be returned.

        """
        if isinstance(func, str):
            func = self.transform_funcs.get(func.lower(), None)
            if func is None:
                func_block = "\n  - ".join(self.transform_funcs.keys())
                err = f"\nERROR: invalid transform function.\n"
                err += f"Options are: {func_block}\n"
                err + +f"You provided: '{func}'\n"
                print(err)
                sys.exit()
        elif not callable(func):
            return
        df = self.raw_df.copy() if use_raw else self.df.copy()

        # subset of the dataframe to transform
        if isinstance(subset, str):
            subset = [subset]
        if subset is None:
            subset = df.columns
        subset = [s for s in subset if s in df.columns]

        # transform
        for c in subset:
            df[c] = func(df[c])

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def clip(
        self,
        lower: Union[int, float, None] = None,
        upper: Union[int, float, None] = None,
        axis: Union[int, str] = "columns",
        subset: Union[Iterable, str, None] = None,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """
        Clip the data. Values less than `lower` will be replaced with `lower` and values
        greater than `upper` will be replaced with `upper`. By default, the entire dataset
        will be clipped. Alternatively, a subset of columns can be clipped.

        Parameters
        ----------
        lower : Union[int, float, None]
            The lower bound. If ``None``, no lower bound will be applied. Default is ``None``.

        upper : Union[int, float, None]
            The upper bound. If ``None``, no upper bound will be applied. Default is ``None``.

        axis : Union[int, str, None]
            The axis to clip. Only used if `subset` is also provided. If ``0``, ``"column"``,
            or ``"columns"``, only column names in `subset` will be clipped. If ``1``,
            ``"row"``, or ``"rows"``, only row names in `subset` will be clipped. Default is
            ``"columns"``.

        subset : Union[Iterable, str, None]
            The columns or rows to clip. Can be either a single column/row name, or an iterable
            of column/row names. If ``None``, all columns/rows will be clipped. Default is
            ``None``.

        use_raw : bool
            If ``True``, the raw data will be clipped. Otherwise, the processed data will be
            clipped. Default is ``False``.

        inplace : bool
            If ``True``, the transformed data will replace ``self.df``. Otherwise, the transformed
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the transformed ``DataFrame`` will be returned.

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        if isinstance(subset, str):
            subset = [subset]

        # clip by column
        if axis in [0, "columns", "column"]:
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = df[c].clip(lower=lower, upper=upper)

        # clip by row
        if axis in [1, "rows", "row"]:
            df = df.T
            subset = subset if subset is not None else df.columns
            subset = [s for s in subset if s in df.columns]
            for c in subset:
                df[c] = df[c].clip(lower=lower, upper=upper)
            df = df.T

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def reorder(
        self,
        by: Union[Iterable, str, None] = None,
        axis: Union[int, str] = "rows",
        row_order: Optional[Iterable] = None,
        column_order: Optional[Iterable] = None,
        ascending: bool = True,
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """ """
        df = self.raw_df.copy() if use_raw else self.df.copy()

        # reorder based on row/column values
        if by is not None:
            if axis in [0, "columns", "column"]:
                df = df.sort_values(by=by, axis=0, ascending=ascending)
            if axis in [1, "rows", "row"]:
                df = df.sort_values(by=by, axis=1, ascending=ascending)

        # reorder based on column names
        if column_order is not None:
            df = df[column_order]

        # reorder based on row names (index)
        if row_order is not None:
            df = df.loc[row_order]

        # update or return
        if inplace:
            self.df = df
        else:
            return df

    def groupby(
        self,
        by: Union[Iterable, str],
        axis: Union[int, str] = "rows",
        values: Optional[Union[Iterable, str]] = None,
        agg: Union[Iterable, Callable, str, dict] = "size",
        use_raw: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """ 
        Group the data by row or column values.

        Parameters
        ----------
        by : Union[Iterable, str]
            The column or row values to group by. Can be either a single column/row name, or an
            iterable of column/row names.

        axis : Union[int, str]
            The axis to group by. If ``0``, ``"column"``, or ``"columns"``, the data will be
            grouped by column values. If ``1``, ``"row"``, or ``"rows"``, the data will be
            grouped by row values. Default is ``"rows"``.

        values : Optional[Union[Iterable, str]]
            The columns or rows to aggregate. Can be either a single column/row name, or an
            iterable of column/row names. If ``None``, all columns/rows will be aggregated.
            Default is ``None``.

        agg : Union[Iterable, Callable, str, dict]
            The aggregation function(s) to apply to the grouped data. Can be a single function,
            an iterable of functions, or a dictionary mapping column/row names to functions.
            Default is ``"size"``.

        use_raw : bool
            If ``True``, the raw data will be grouped. Otherwise, the processed data will be
            grouped. Default is ``False``.

        inplace : bool
            If ``True``, the transformed data will replace ``self.df``. Otherwise, the transformed
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the transformed ``DataFrame`` will be returned.

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        # do the groupby
        if axis in [1, "rows", "row"]:
            g = df.groupby(by=by, axis=1)
        else:
            g = df.groupby(by=by, axis=0)
        # filter values
        if values is not None:
            g = g[values]
        # aggregate
        agg_df = g.agg(agg)
        if inplace:
            self.df = agg_df
        else:
            return agg_df

    def to_squareform(
        self,
        rows: Optional[str] = None,
        columns: Optional[str] = None,
        values: Optional[str] = None,
        agg: Union[Iterable, Callable, str] = "size",
        use_raw: bool = False,
        reset_index: bool = False,
        inplace: bool = True,
    ) -> Optional[pd.DataFrame]:
        """ 
        Pivots the data into a squareform matrix.

        Parameters
        ----------
        rows : Optional[str]
            The column name to use for the rows. If ``None``, ``self.x_name`` will be used. 
            Default is ``None``.
        
        columns : Optional[str]
            The column name to use for the columns. If ``None``, ``self.y_name`` will be used. 
            Default is ``None``.

        values : Optional[str]
            The column name to use for the aggregated values. If ``None``, all columns will 
            be used. Default is ``None``.

        agg : Union[Iterable, Callable, str]
            The aggregation function(s) to use. If a single function is provided, it will be
            applied to all values. If multiple functions are provided, they will be applied
            to the corresponding values. If a string is provided, it must be a valid aggregation
            function. Default is ``"size"``.

        use_raw : bool
            If ``True``, the raw data will be used. Otherwise, the processed data will be used.
            Default is ``False``.

        reset_index : bool
            If ``True``, the index will be reset. Default is ``False``.

        inplace : bool
            If ``True``, the transformed data will replace ``self.df``. Otherwise, the transformed
            ``DataFrame`` will be returned. Default is ``True``.

        Returns
        -------
        Optional[pd.DataFrame]
            If ``inplace`` is ``False``, the transformed ``DataFrame`` will be returned.

        """
        df = self.raw_df.copy() if use_raw else self.df.copy()
        # get aggregation function(s)
        if isinstance(agg, str) or callable(agg):
            agg = [agg]
        agg_funcs = []
        for a in agg:
            if isinstance(a, str):
                a = self.agg_funcs.get(a.lower(), None)
                if a is None:
                    agg_block = "\n  - ".join(self.agg_funcs.keys())
                    err = f"\nERROR: invalid aggregation function.\n"
                    err += f"Options are: {agg_block}\n"
                    err + +f"You provided: '{a}'\n"
                    print(err)
                    sys.exit()
            elif not callable(a):
                agg_block = "\n  - ".join(self.agg_funcs.keys())
                err = f"\nERROR: invalid aggregation function.\n"
                err += f"<agg> must be a function or the name of a built-in aggregation function.\n"
                err += f"Built-in options are: {agg_block}\n"
                err + +f"You provided: '{a}'\n"
                print(err)
                sys.exit()
            agg_funcs.append(a)
        agg_funcs = [a for a in agg_funcs if a is not None]
        if len(agg_funcs) == 1:
            agg_funcs = agg_funcs[0]

        # group by columns/rows
        if columns is None:
            columns = self.x_name
        if rows is None:
            rows = self.y_name
        g = df.groupby([rows, columns])

        # aggregate
        if values is not None:
            agg_df = g[values].agg(agg_funcs)
        else:
            agg_df = g.agg(agg_funcs)

        # pivot to squareform
        sq_df = agg_df.unstack()
        if reset_index:
            sq_df.reset_index(inplace=True)

        # update or return
        if inplace:
            self.df = sq_df
        else:
            return sq_df



# from .core import *
# from .utils import *

import os

from . import bin, cl, io, log, pl, tl
from . import cl as color
from .core import lineage, pair, sequence
from .core.lineage import Lineage
from .core.pair import Pair
from .core.sequence import Sequence

# from .tools import log
from .tools.phylo import Phylogeny
from .utils import (
    alignment,
    cluster,
    # color,
    decorators,
    jobs,
    # log,
    phylogeny,
    pipeline,
    progbar,
    s3,
    seqio,
    utilities,
)
from .utils import alignment as aln
from .utils import pipeline as path

# from .utils.alignment import SSWAlignment, NWAlignment
from .utils.alignment import GlobalAlignment, LocalAlignment, SemiGlobalAlignment
from .utils.jobs import *
from .utils.progbar import progress_bar
from .version import __version__

BINARY_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "binaries"))



#!/usr/bin/env python
# filename: pair.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import csv
import os
from typing import Callable, Iterable, Optional, Union

import pandas as pd
import polars as pl

from ..core.sequence import Sequence

# from ..utils.utilities import nested_dict_lookup


class Pair(object):
    """
    Holds a one or more ``Sequence`` objects, corresponding to a paired mAb or TCR.

    """

    def __init__(
        self,
        sequences: Iterable[Union[dict, Sequence]],
        name: Optional[str] = None,
        chain_selection_func: Optional[Callable] = None,
        properties: Optional[dict] = None,
    ):
        """
        Initialize a Pair object.

        Parameters
        ----------
        sequences : Iterable[Union[dict, Sequence]]
            A list of sequence objects, each containing sequence information.

        name : Optional[str], default=None
            The name of the pair.

        chain_selection_func : Optional[Callable], default=None
            A function that takes a list of sequences and orders them to determine
            the "correct" heavy and light chains in cases for which multiple heavy
            or light chains exist. If not provided, chains are prioritized in the order
            provided.

        properties : Optional[dict], default=None
            A dictionary of additional properties to add to the Pair object.

        """
        self._sequences = sequences
        self._receptor = None
        self._heavy = None
        self._light = None
        self._alpha = None
        self._beta = None
        self._delta = None
        self._gamma = None
        self._heavies = None
        self._lights = None
        self._alphas = None
        self._betas = None
        self._deltas = None
        self._gammas = None
        self._name = name
        self._fasta = None
        self._sample = None
        self._subject = None
        self._group = None
        self._experiment = None
        self._timepoint = None
        self._is_pair = None
        self._lineage = None
        self._select_chain = (
            chain_selection_func
            if chain_selection_func is not None
            else self._chain_selector
        )
        if properties is not None:
            for k, v in properties.items():
                setattr(self, k, v)

    def __eq__(self, other):
        return (self.heavy, self.light) == (other.heavy, other.light)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        name = self.name
        heavy = self.heavy.sequence if self.heavy is not None else None
        light = self.light.sequence if self.light is not None else None
        return hash((name, heavy, light))

    @property
    def receptor(self):
        if self._receptor is None:
            # check for the "locus" annotation
            if all([s["locus"] is not None for s in self._sequences]):
                if all(
                    [
                        s["locus"].upper() in ["IGH", "IGK", "IGL"]
                        for s in self._sequences
                    ]
                ):
                    self._receptor = "bcr"
                elif all(
                    [
                        s["locus"].upper() in ["TRA", "TRB", "TRD", "TRG"]
                        for s in self._sequences
                    ]
                ):
                    self._receptor = "tcr"
                else:
                    self._receptor = "unknown"

            # if locus isn't present, infer from V-gene assignment
            elif all([s["v_call"] is not None for s in self._sequences]):
                if all(
                    [
                        s["v_call"][:3].upper() in ["IGH", "IGK", "IGL"]
                        for s in self._sequences
                    ]
                ):
                    self._receptor = "bcr"
                elif all(
                    [
                        s["v_call"][:3].upper() in ["TRA", "TRB", "TRD", "TRG"]
                        for s in self._sequences
                    ]
                ):
                    self._receptor = "tcr"
                else:
                    self._receptor = "unknown"
        return self._receptor

    @receptor.setter
    def receptor(self, receptor):
        self._receptor = receptor

    @property
    def heavy(self):
        if self._heavy is None:
            self._heavy = self._select_chain(self.heavies)
        return self._heavy

    @heavy.setter
    def heavy(self, heavy):
        self._heavy = heavy

    @property
    def light(self):
        if self._light is None:
            self._light = self._select_chain(self.lights)
        return self._light

    @light.setter
    def light(self, light):
        self._light = light

    @property
    def alpha(self):
        if self._alpha is None:
            self._alpha = self._select_chain(self.alphas)
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha

    @property
    def beta(self):
        if self._beta is None:
            self._beta = self._select_chain(self.betas)
        return self._beta

    @beta.setter
    def beta(self, beta):
        self._beta = beta

    @property
    def delta(self):
        if self._delta is None:
            self._delta = self._select_chain(self.deltas)
        return self._delta

    @delta.setter
    def delta(self, delta):
        self._delta = delta

    @property
    def gamma(self):
        if self._gamma is None:
            self._gamma = self._select_chain(self.gammas)
        return self._gamma

    @gamma.setter
    def gamma(self, gamma):
        self._gamma = gamma

    @property
    def heavies(self):
        if self._heavies is None:
            self._heavies = [
                s for s in self._sequences if self._is_locus_type(s, ["IGH", "heavy"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._heavies = [s for s in self._seqs if s["chain"] == "heavy"]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._heavies = [s for s in self._seqs if s["locus"] == "IGH"]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._heavies = [
            #         s for s in self._seqs if s["v_gene"][:3].upper() == "IGH"
            #     ]
        return self._heavies

    @property
    def lights(self):
        if self._lights is None:
            self._lights = [
                s
                for s in self._sequences
                if self._is_locus_type(s, ["IGK", "IGL", "kappa", "lambda"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._lights = [
            #         s for s in self._seqs if s["chain"] in ["kappa", "lambda"]
            #     ]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._lights = [s for s in self._seqs if s["locus"] in ["IGK", "IGL"]]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._lights = [
            #         s for s in self._seqs if s["v_gene"][:3].upper() in ["IGK", "IGL"]
            #     ]
        return self._lights

    @property
    def alphas(self):
        if self._alphas is None:
            self._alphas = [
                s for s in self._sequences if self._is_locus_type(s, ["TRA", "alpha"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._alphas = [s for s in self._seqs if s["chain"] == "alpha"]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._alphas = [s for s in self._seqs if s["locus"] == "TRA"]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._alphas = [
            #         s for s in self._seqs if s["v_gene"][:3].upper() == "TRA"
            #     ]
        return self._alphas

    @property
    def betas(self):
        if self._betas is None:
            self._betas = [
                s for s in self._sequences if self._is_locus_type(s, ["TRB", "beta"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._betas = [s for s in self._seqs if s["chain"] == "beta"]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._betas = [s for s in self._seqs if s["locus"] == "TRB"]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._betas = [
            #         s for s in self._seqs if s["v_gene"][:3].upper() == "TRB"
            #     ]
        return self._betas

    @property
    def deltas(self):
        if self._deltas is None:
            self._deltas = [
                s for s in self._sequences if self._is_locus_type(s, ["TRD", "delta"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._deltas = [s for s in self._seqs if s["chain"] == "delta"]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._deltas = [s for s in self._seqs if s["locus"] == "TRD"]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._deltas = [
            #       s for s in self._seqs if s["v_gene"][:3].upper() == "TRD"
            #     ]
        return self._deltas

    @property
    def gammas(self):
        if self._gammas is None:
            self._gammas = [
                s for s in self._sequences if self._is_locus_type(s, ["TRG", "gamma"])
            ]
            # if all([s["chain"] is not None for s in self._seqs]):
            #     self._gammas = [s for s in self._seqs if s["chain"] == "gamma"]
            # elif all([s["locus"] is not None for s in self._seqs]):
            #     self._gammas = [s for s in self._seqs if s["locus"] == "TRG"]
            # elif all([s["v_gene"] is not None for s in self._seqs]):
            #     self._gammas = [
            #         s for s in self._seqs if s["v_gene"][:3].upper() == "TRG"
            #     ]
        return self._gammas

    @property
    def is_pair(self):
        if all([self.heavy is not None, self.light is not None]):
            return True
        elif all([self.alpha is not None, self.beta is not None]):
            return True
        elif all([self.gamma is not None, self.delta is not None]):
            return True
        return False

    @property
    def lineage(self):
        if self._lineage is None:
            if "clonify" in self.heavy:
                self._lineage = self.heavy["clonify"]["id"]
        return self._lineage

    @lineage.setter
    def lineage(self, lineage):
        self._lineage = lineage

    @property
    def name(self):
        if self._name is None:
            if self.heavy is not None:
                if self.heavy.id is not None:
                    self._name = self.heavy.id
                elif self.heavy["sequence_id"] is not None:
                    self._name = self.heavy["sequence_id"]
                # elif self.heavy["seq_id"] is not None:
                #     self._name = self.heavy["seq_id"]
                else:
                    pass
            elif self.light is not None:
                if self.light.id is not None:
                    self._name = self.light.id
                elif self.light["sequence_id"] is not None:
                    self._name = self.light["sequence_id"]
                # elif self.light["seq_id"] is not None:
                #     self._name = self.light["seq_id"]
                else:
                    pass
            elif self.alpha is not None:
                if self.alpha.id is not None:
                    self._name = self.alpha.id
                elif self.alpha["sequence_id"] is not None:
                    self._name = self.alpha["sequence_id"]
                # elif self.alpha["seq_id"] is not None:
                #     self._name = self.alpha["seq_id"]
                else:
                    pass
            elif self.beta is not None:
                if self.beta.id is not None:
                    self._name = self.beta.id
                elif self.beta["sequence_id"] is not None:
                    self._name = self.beta["sequence_id"]
                # elif self.beta["seq_id"] is not None:
                #     self._name = self.beta["seq_id"]
                else:
                    pass
            elif self.delta is not None:
                if self.delta.id is not None:
                    self._name = self.delta.id
                elif self.delta["sequence_id"] is not None:
                    self._name = self.delta["sequence_id"]
                # elif self.delta["seq_id"] is not None:
                #     self._name = self.delta["seq_id"]
                else:
                    pass
            elif self.gamma is not None:
                if self.gamma.id is not None:
                    self._name = self.gamma.id
                elif self.gamma["sequence_id"] is not None:
                    self._name = self.gamma["sequence_id"]
                # elif self.gamma["seq_id"] is not None:
                #     self._name = self.gamma["seq_id"]
                else:
                    pass
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def _is_locus_type(self, seq: Sequence, locus_names: Iterable[str]) -> bool:
        """
        Check if the sequence is of the specified locus type.

        Parameters
        ----------
        seq : Sequence
            The sequence to check.

        locus_names : list[str]
            The list of locus names to check against.

        Returns
        -------
        bool
            True if the sequence is of the specified locus type, False otherwise.
        """
        if seq["locus"] in locus_names:
            return True
        elif seq["v_call"] is not None:
            if seq["v_call"][:3] in locus_names:
                return True
        return False

    # @property
    # def sample(self):
    #     if self._sample is None:
    #         slist = []
    #         if self.experiment is not None:
    #             slist.append(str(self.experiment))
    #         if self.group is not None:
    #             slist.append(str(self.group))
    #         if self.subject is not None:
    #             slist.append(str(self.subject))
    #         if self.timepoint is not None:
    #             slist.append(str(self.timepoint))
    #         if slist:
    #             self._sample = '|'.join(slist)
    #     return self._sample

    # @property
    # def subject(self):
    #     if self._subject is None:
    #         if self.heavy is not None and 'subject' in self.heavy.keys():
    #             self._subject = self.heavy['subject']
    #         elif self.light is not None and 'subject' in self.light.keys():
    #             self._subject = self.light['subject']
    #     return self._subject

    # @subject.setter
    # def subject(self, subject):
    #     self._subject = subject

    # @property
    # def group(self):
    #     if self._group is None:
    #         if self.heavy is not None and 'group' in self.heavy.keys():
    #             self._group = self.heavy['group']
    #         elif self.light is not None and 'group' in self.light.keys():
    #             self._group = self.light['group']
    #     return self._group

    # @group.setter
    # def group(self, group):
    #     self._group = group

    # @property
    # def experiment(self):
    #     if self._experiment is None:
    #         if self.heavy is not None and 'experiment' in self.heavy.keys():
    #             self._experiment = self.heavy['experiment']
    #         elif self.light is not None and 'experiment' in self.light.keys():
    #             self._experiment = self.light['experiment']
    #     return self._experiment

    # @experiment.setter
    # def experiment(self, experiment):
    #     self._experiment = experiment

    # @property
    # def timepoint(self):
    #     if self._timepoint is None:
    #         if self.heavy is not None and 'timepoint' in self.heavy.keys():
    #             self._timepoint = self.heavy['timepoint']
    #         elif self.light is not None and 'timepoint' in self.light.keys():
    #             self._timepoint = self.light['timepoint']
    #     return self._timepoint

    # @timepoint.setter
    # def timepoint(self, timepoint):
    #     self._timepoint = timepoint

    # def refine(self, heavy=True, light=True, species='human'):
    #     for seq in [s for s in [self.heavy, self.light] if s is not None]:
    #         try:
    #             self.remove_ambigs(seq)
    #             self._refine_v(seq, species)
    #             self._refine_j(seq, species)
    #             self._retranslate(seq)
    #         except:
    #             print('REFINEMENT FAILED: {}, {} chain'.format(seq['seq_id'], seq['chain']))
    #             print(traceback.format_exception_only(*sys.exc_info()[:2]))

    # @staticmethod
    # def remove_ambigs(seq):
    #     # fix Ns in the nucleotide sequence
    #     vdj = ''
    #     for s, g in zip(seq['vdj_nt'], seq['vdj_germ_nt']):
    #         if s.upper() == 'N':
    #             vdj += g
    #         else:
    #             vdj += s
    #     seq['vdj_nt'] = vdj
    #     # fix Xs in the amino acid sequence
    #     vdj = ''
    #     for s, g in zip(seq['vdj_aa'], seq['vdj_germ_aa']):
    #         if s.upper() == 'X':
    #             vdj += g
    #         else:
    #             vdj += s
    #     seq['vdj_aa'] = vdj

    # @staticmethod
    # def _refine_v(seq, species):
    #     '''
    #     Completes the 5' end of a a truncated sequence with germline nucleotides.
    #     Input is a MongoDB dict (seq) and the species.
    #     '''
    #     vgerm = germlines.get_germline(seq['v_gene']['full'], species)
    #     aln = global_alignment(seq['vdj_nt'], vgerm)
    #     prepend = ''
    #     for s, g in zip(aln.aligned_query, aln.aligned_target):
    #         if s != '-':
    #             break
    #         else:
    #             prepend += g
    #     seq['vdj_nt'] = prepend + seq['vdj_nt']

    # @staticmethod
    # def _refine_j(seq, species):
    #     '''
    #     Completes the 3' end of a a truncated sequence with germline nucleotides.
    #     Input is a MongoDB dict (seq) and the species.
    #     '''
    #     jgerm = germlines.get_germline(seq['j_gene']['full'], species)
    #     aln = global_alignment(seq['vdj_nt'], jgerm)
    #     append = ''
    #     for s, g in zip(aln.aligned_query[::-1], aln.aligned_target[::-1]):
    #         if s != '-':
    #             break
    #         else:
    #             append += g
    #     seq['vdj_nt'] = seq['vdj_nt'] + append[::-1]

    # @staticmethod
    # def _retranslate(seq):
    #     '''
    #     Retranslates a nucleotide sequence following refinement.
    #     Input is a Pair sequence (basically a dict of MongoDB output).
    #     '''
    #     if len(seq['vdj_nt']) % 3 != 0:
    #         trunc = len(seq['vdj_nt']) % 3
    #         seq['vdj_nt'] = seq['vdj_nt'][:-trunc]
    #     seq['vdj_aa'] = Seq(seq['vdj_nt']).translate()

    def fasta(
        self,
        name_field: str = "sequence_id",
        sequence_field: str = "sequence",
        append_chain: bool = True,
    ):
        """
        Returns the sequence pair as a fasta string. If the Pair object contains
        both heavy and light chain sequences, both will be returned as a single string.

        By default, the fasta string contains the 'vdj_nt' sequence for each chain. To change,
        use the <key> option to select an alternate sequence.

        By default, the chain (heavy or light) will be appended to the sequence name:

        >MySequence_heavy

        To just use the pair name (which will result in duplicate sequence names for Pair objects
        with both heavy and light chains), set <append_chain> to False.
        """
        fastas = []
        for s, chain in [(self.heavy, "heavy"), (self.light, "light")]:
            if s is not None:
                c = "_{}".format(chain) if append_chain else ""
                fastas.append(">{}{}\n{}".format(s[name_field], c, s[sequence_field]))
        return "\n".join(fastas)

    def summarize(self, annotation_format: Optional[str] = None) -> None:
        """
        Prints a summary of the pair to the console.

        Parameters
        ----------
        annotation_format : str, optional
            No longer used, retained for legacy compatibility.

        Returns
        -------
        None
        """
        try:
            vline = ""
            dline = ""
            jline = ""
            juncline = ""
            identline = ""
            isotypeline = ""
            if self.heavy is not None:
                vline += self.heavy["v_gene"]
                dline += self.heavy["d_gene"]
                jline += self.heavy["j_gene"]
                juncline += self.heavy["junction_aa"]
                identline += self.heavy["v_identity"]
                isotypeline += self.heavy["isotype"]
            pad = (
                max(
                    [
                        len(line)
                        for line in [
                            vline,
                            dline,
                            jline,
                            juncline,
                            identline,
                            isotypeline,
                        ]
                    ]
                )
                + 4
            )
            vline += " " * (pad - len(vline))
            dline += " " * (pad - len(dline))
            jline += " " * (pad - len(jline))
            juncline += " " * (pad - len(juncline))
            identline += " " * (pad - len(identline))
            isotypeline += " " * (pad - len(isotypeline))
            if self.light is not None:
                vline += self.light["v_gene"]
                jline += self.light["j_gene"]
                juncline += self.light["junction_aa"]
                identline += self.light["v_identity"]
                isotypeline += self.light["isotype"]

            header_len = max(
                [
                    len(line)
                    for line in [vline, dline, jline, juncline, identline, isotypeline]
                ]
            )
            header_spaces = int((header_len - len(self.name)) / 2)
            print(f"{' ' * header_spaces}{self.name}")
            print("-" * header_len)
            for line in [vline, dline, jline, juncline, identline, isotypeline]:
                print(line)
            print("")
        except Exception:
            return

    def _chain_selector(self, seqs):
        if len(seqs) == 0:
            return None
        if all(["umis" in s for s in seqs]):
            sorted_seqs = sorted(seqs, key=lambda x: x["umis"], reverse=True)
            return sorted_seqs[0]
        else:
            return seqs[0]


def get_pairs(
    db,
    collection,
    experiment=None,
    subject=None,
    group=None,
    name="seq_id",
    delim=None,
    delim_occurance=1,
    pairs_only=False,
    chain_selection_func=None,
):
    """
    Gets sequences and assigns them to the appropriate mAb pair, based on the sequence name.

    Inputs:

    ::db:: is a pymongo database connection object
    ::collection:: is the collection name, as a string
    If ::subject:: is provided, only sequences with a 'subject' field matching ::subject:: will
        be included. ::subject:: can be either a single subject (as a string) or an iterable
        (list or tuple) of subject strings.
    If ::group:: is provided, only sequences with a 'group' field matching ::group:: will
        be included. ::group:: can be either a single group (as a string) or an iterable
        (list or tuple) of group strings.
    ::name:: is the dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'
    ::delim:: is an optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.
    ::delim_occurance:: is the occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.
    ::pairs_only:: setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    Returns a list of Pair objects, one for each mAb pair.
    """
    match = {}
    if subject is not None:
        if type(subject) in (list, tuple):
            match["subject"] = {"$in": subject}
        elif isinstance(subject, str):
            match["subject"] = subject
    if group is not None:
        if type(group) in (list, tuple):
            match["group"] = {"$in": group}
        elif isinstance(group, str):
            match["group"] = group
    if experiment is not None:
        if type(experiment) in (list, tuple):
            match["experiment"] = {"$in": experiment}
        elif isinstance(experiment, str):
            match["experiment"] = experiment
    seqs = list(db[collection].find(match))
    return assign_pairs(
        seqs,
        name=name,
        delim=delim,
        delim_occurance=delim_occurance,
        pairs_only=pairs_only,
        chain_selection_func=chain_selection_func,
    )


def assign_pairs(
    seqs: Iterable[Union[dict, Sequence]],
    id_key: str = "sequence_id",
    delim: Optional[str] = None,
    delim_occurance: int = 1,
    pairs_only: bool = False,
    chain_selection_func: Optional[Callable] = None,
    tenx_annot_file: Optional[str] = None,
) -> Iterable[Pair]:
    """
    Assigns sequences to the appropriate mAb pair, based on the sequence name.

    Parameters
    ----------
    seqs : Iterable[Union[dict, Sequence]]
        List of sequence objects, of the format returned by querying a MongoDB containing
        Abstar output.

    id_key : str, default="sequence_id"
        The dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'

    delim : Optional[str], default=None
        An optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.

    delim_occurance : int, default=1
        The occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.

    pairs_only : bool, default=False
        Setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    chain_selection_func : Optional[Callable], default=None
        A function that takes a list of sequences and returns a single sequence.
        Default is None, which results in the first sequence in the list being returned.

    tenx_annot_file : Optional[str], default=None
        A path to a 10x Genomics annotations file. If provided, the UMIs and 10x
        annotations will be added to the sequence annotations.

    Returns
    -------
    pairs : Iterable[Pair]
        A list of ``Pair`` objects, one for each mAb pair.

    """
    # add UMIs to the sequence annotations if a 10xG annotations file is provided
    if tenx_annot_file is not None:
        annots = {}
        umis = {}
        with open(tenx_annot_file) as f:
            reader = csv.DictReader(f)
            for r in reader:
                key = "consensus_id" if "consensus_id" in r else "contig_id"
                umis[r[key]] = r["umis"]
                annots[r[key]] = r
        for s in seqs:
            s["umis"] = int(umis.get(s[id_key], 0))
            s["tenx_annots"] = annots.get(s[id_key], {})

    # identify pairs
    pdict = {}
    for s in seqs:
        if delim is not None:
            pname = delim.join(s[id_key].split(delim)[:delim_occurance])
        else:
            pname = s[id_key]
        if pname not in pdict:
            pdict[pname] = [
                s,
            ]
        else:
            pdict[pname].append(s)
    pairs = [
        Pair(p, name=n, chain_selection_func=chain_selection_func)
        for n, p in pdict.items()
    ]
    if pairs_only:
        pairs = [p for p in pairs if p.is_pair]
    return pairs


def pairs_to_polars(
    pairs: Iterable[Pair],
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> pl.DataFrame:
    """
    Converts a list of ``Pair`` objects to a Polars DataFrame.

    Parameters
    ----------
    pairs : Iterable[Pair]
        List of ``Pair`` objects to be converted to a Polars DataFrame. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the Polars DataFrame. The difference
        between ``annotations`` and ``columns`` is that ``annotations`` refers to fields
        in the heavy/light chain annotations (like ``"sequence"``, ``"v_gene"``, etc.),
        while ``columns`` refers to fields in the Polars DataFrame (like ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc.).

    columns : list, default=None
        A list of fields to be retained in the output Polars DataFrame. Fields must be column
        names in the input file. The difference between ``annotations`` and ``columns`` is
        that ``annotations`` refers to fields in the heavy/light chain annotations (like
        ``"sequence"``, ``"v_gene"``, etc.), while ``columns`` refers to fields in the
        Polars DataFrame (like ``"sequence:0"``, ``"sequence:1"``, ``"name"``, etc.).

    properties : list, default=None
        A list of properties to be included in the Polars DataFrame. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the Polars DataFrame.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the Polars DataFrame.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the Polars DataFrame.

    leading : str or list, default=None
        Field or list of fields to appear first in the Polars DataFrame. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the Polars DataFrame and
        remaining fields will appear in the order provided in ``order``.

    """
    data = []
    # read Pair data
    for p in pairs:
        d = {"name": p.name}
        # heavy chain
        if p.heavy is not None:
            if not p.heavy.annotations:
                d.update({"sequence_id:0": p.heavy.id, "sequence:0": p.heavy.sequence})
            else:
                if annotations is not None:
                    keys = [a for a in annotations if a in p.heavy.annotations]
                else:
                    keys = list(p.heavy.annotations.keys())
                d.update({f"{k}:0": p.heavy[k] for k in keys})
            if sequence_properties is not None:
                for prop in sequence_properties:
                    try:
                        d[f"{prop}:0"] = getattr(p.heavy, prop)
                    except AttributeError:
                        pass
        # light chain
        if p.light is not None:
            if not p.light.annotations:
                d.update({"sequence_id:1": p.light.id, "sequence:1": p.light.sequence})
            else:
                if annotations is not None:
                    keys = [a for a in annotations if a in p.light.annotations]
                else:
                    keys = list(p.light.annotations.keys())
                d.update({f"{k}:1": p.light[k] for k in keys})
            if sequence_properties is not None:
                for prop in sequence_properties:
                    try:
                        d[f"{prop}:1"] = getattr(p.light, prop)
                    except AttributeError:
                        pass
        # properties
        if properties is not None:
            for prop in properties:
                try:
                    d[prop] = getattr(p, prop)
                except AttributeError:
                    continue
        data.append(d)

    # build dataframe
    df = pl.DataFrame(data, infer_schema_length=len(data))

    # drop NaN
    if drop_na_columns:
        df = df[[s.name for s in df if not (s.null_count() == df.height)]]

    # excluded columns
    if exclude is not None:
        if isinstance(exclude, str):
            exclude = []
        cols = [c for c in df.columns if c not in exclude]
        df = df.select(cols)

    # reorder
    if order is not None:
        cols = [o for o in order if o in df.columns]
        df = df.select(cols)

    # leading columns
    if leading is not None:
        if isinstance(leading, str):
            leading = [leading]
        leading = [l for l in leading if l in df.columns]
        cols = leading + [c for c in df.columns if c not in leading]
        df = df.select(cols)

    if columns is not None:
        cols = [c for c in columns if c in df.columns]
        df = df.select(cols)

    return df


def pairs_from_polars(
    df: Union[pl.DataFrame, pl.LazyFrame],
    match: Optional[dict] = None,
    fields: Iterable = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
) -> Iterable[Pair]:
    """
    Reads a Polars DataFrame and returns ``Pair`` objects.

    Parameters
    ----------
    df : Union[pl.DataFrame, pl.LazyFrame]
        The input Polars DataFrame. Required.

    match : dict, default=None
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, default=None
        A ``list`` of fields to be retained in the output ``Pair``
        objects. Fields must be column names in the input file.

    id_key : str, default="sequence_id"
        Name of the annotation field containing the sequence ID. Used to
        populate the ``Sequence.id`` property.

    sequence_key : str, default="sequence"
        Name of the annotation field containg the sequence. Used to
        populate the ``Sequence.sequence`` property.


    Returns
    -------
    pairs : list of ``Pairs``

    """
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    pairs = []
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    for r in df.iter_rows(named=True):
        try:
            if all([r[k] == v for k, v in match.items()]):
                heavy = None
                light = None
                name = r.get("name", None)
                # heavy chain
                heavy_dict = {
                    k.split(":")[0]: v for k, v in r.items() if k.endswith(":0")
                }
                if fields is not None:
                    hc_fields = [f for f in fields if f in heavy_dict]
                    heavy_dict = {f: heavy_dict[f] for f in hc_fields}
                if any([v is not None for v in heavy_dict.values()]):
                    heavy = Sequence(heavy_dict, id_key=id_key, seq_key=sequence_key)
                # light chain
                light_dict = {
                    k.split(":")[0]: v for k, v in r.items() if k.endswith(":1")
                }
                if fields is not None:
                    lc_fields = [f for f in fields if f in light_dict]
                    light_dict = {f: light_dict[f] for f in lc_fields}
                if any([v is not None for v in light_dict.values()]):
                    light = Sequence(light_dict, id_key=id_key, seq_key=sequence_key)
                # extra fields:
                extra_fields = {}
                for k, v in r.items():
                    if (
                        not k.endswith(":0")
                        and not k.endswith(":1")
                        and k.lower() != "name"
                    ):
                        extra_fields[k] = v
                # pair
                sequences = [s for s in [heavy, light] if s is not None]
                pairs.append(
                    Pair(sequences=sequences, name=name, properties=extra_fields)
                )
        except KeyError:
            continue
    return pairs


def pairs_to_pandas(
    pairs: Iterable[Pair],
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> pd.DataFrame:
    """
    Saves a list of ``Pair`` objects to a CSV file.

    Parameters
    ----------
    pairs : Iterable[Pair]
        List of ``Pair`` objects to be saved to a CSV file. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the Polars DataFrame. The difference
        between ``annotations`` and ``columns`` is that ``annotations`` refers to fields
        in the heavy/light chain annotations (like ``"sequence"``, ``"v_gene"``, etc.),
        while ``columns`` refers to fields in the Polars DataFrame (like ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc.).

    columns : list, default=None
        A list of fields to be retained in the output Polars DataFrame. Fields must be column
        names in the input file. The difference between ``annotations`` and ``columns`` is
        that ``annotations`` refers to fields in the heavy/light chain annotations (like
        ``"sequence"``, ``"v_gene"``, etc.), while ``columns`` refers to fields in the
        Polars DataFrame (like ``"sequence:0"``, ``"sequence:1"``, ``"name"``, etc.).

    properties : list, default=None
        A list of properties to be included in the CSV file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the CSV file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the CSV file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the CSV file.

    leading : str or list, default=None
        Field or list of fields to appear first in the CSV file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the CSV file and
        remaining fields will appear in the order provided in ``order``.

    """
    df = pairs_to_polars(
        pairs,
        annotations=annotations,
        columns=columns,
        properties=properties,
        sequence_properties=sequence_properties,
        drop_na_columns=drop_na_columns,
        order=order,
        exclude=exclude,
        leading=leading,
    )
    return df.to_pandas()


def pairs_from_pandas(
    df: pd.DataFrame,
    match: Optional[dict] = None,
    fields: Iterable = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
) -> Iterable[Pair]:
    """
    Reads a Polars DataFrame and returns ``Pair`` objects.

    Parameters
    ----------
    df : pd.DataFrame
        The input pandas DataFrame. Required.

    match : dict, default=None
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, default=None
        A ``list`` of fields to be retained in the output ``Pair``
        objects. Fields must be column names in the input file.

    id_key : str, default="sequence_id"
        Name of the annotation field containing the sequence ID. Used to
        populate the ``Sequence.id`` property.

    sequence_key : str, default="sequence"
        Name of the annotation field containg the sequence. Used to
        populate the ``Sequence.sequence`` property.


    Returns
    -------
    pairs : list of ``Pairs``

    """
    polars_df = pl.from_pandas(df)
    return pairs_from_polars(
        df=polars_df,
        match=match,
        fields=fields,
        id_key=id_key,
        sequence_key=sequence_key,
    )


def pairs_to_csv(
    pairs: Iterable[Pair],
    csv_file: Optional[str] = None,
    separator: str = ",",
    header: bool = True,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> None:
    """
    Saves a list of ``Pair`` objects to a CSV file.

    Parameters
    ----------
    pairs : Iterable[Pair]
        List of ``Pair`` objects to be saved to a CSV file. Required.

    csv_file : str
        Path to the output CSV file. If not provided, a ``polars.DataFrame``
        containing the CSV data will be returned.

    separator : str, default=","
        Column separator. Default is ``","``.

    header : bool, default=True
        If ``True``, the CSV file will contain a header row. Default is ``True``.

    columns : list, default=None
        A list of fields to be retained in the output CSV file. Fields must be column
        names in the input file.

    properties : list, default=None
        A list of properties to be included in the CSV file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the CSV file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the CSV file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the CSV file.

    leading : str or list, default=None
        Field or list of fields to appear first in the CSV file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the CSV file and
        remaining fields will appear in the order provided in ``order``.

    """
    df = pairs_to_polars(
        pairs,
        columns=columns,
        properties=properties,
        sequence_properties=sequence_properties,
        drop_na_columns=drop_na_columns,
        order=order,
        exclude=exclude,
        leading=leading,
    )
    df.write_csv(csv_file, separator=separator, include_header=header)


def pairs_to_parquet(
    pairs: Iterable[Pair],
    parquet_file: Optional[str] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> None:
    """
    Saves a list of ``Pair`` objects to a Parquet file.

    Parameters
    ----------
    pairs : Iterable[Pair]
        List of ``Pair`` objects to be saved to a Parquet file. Required.

    parquet_file : str
        Path to the output Parquet file. If not provided, a ``polars.DataFrame``
        containing the Parquet data will be returned.

    columns : list, default=None
        A list of fields to be retained in the output Parquet file. Fields must be column
        names in the input file.

    properties : list, default=None
        A list of properties to be included in the Parquet file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the Parquet file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the Parquet file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the Parquet file.

    leading : str or list, default=None
        Field or list of fields to appear first in the Parquet file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the Parquet file and
        remaining fields will appear in the order provided in ``order``.

    """
    df = pairs_to_polars(
        pairs,
        columns=columns,
        properties=properties,
        sequence_properties=sequence_properties,
        drop_na_columns=drop_na_columns,
        order=order,
        exclude=exclude,
        leading=leading,
    )
    df.write_parquet(parquet_file)


# def pairs_to_fasta(
#     pairs: Iterable[Pair],
#     fasta_file: Optional[str] = None,
#     id_key: str = "sequence_id",
#     sequence_key: str = "sequence",
#     append_chain: bool = False,
# ) -> str:
#     """
#     Saves a list of ``Pair`` objects to a FASTA-formatted file.

#     Parameters
#     ----------
#     pairs : Iterable[Pair]
#         List of ``Pair`` objects to be saved to a FASTA-formatted file. Required.

#     fasta_file : str
#         Path to the output FASTA-formatted file. If not provided, the FASTA-formatted
#         sequences will be returned as a string.

#     Returns
#     -------
#     fasta_str : str
#         FASTA-formatted sequences as a string if ``fasta_file`` is not provided. If
#         ``fasta_file`` is provided, the path to the output file is returned.

#     """
#     fastas = [
#         p.to_fasta(id_key=id_key, sequence_key=sequence_key, append_chain=append_chain)
#         for p in pairs
#     ]
#     if fasta_file is None:
#         return "\n".join(fastas)
#     else:
#         from ..io import make_dir

#         make_dir(os.path.dirname(fasta_file))
#         with open(fasta_file, "w") as f:
#             f.write("\n".join(fastas))
#         return fasta_file


# def deduplicate(pairs, aa=False, ignore_primer_regions=False):
#     '''
#     Removes duplicate sequences from a list of Pair objects.

#     If a Pair has heavy and light chains, both chains must identically match heavy and light chains
#     from another Pair to be considered a duplicate. If a Pair has only a single chain,
#     identical matches to that chain will cause the single chain Pair to be considered a duplicate,
#     even if the comparison Pair has both chains.

#     Note that identical sequences are identified by simple string comparison, so sequences of
#     different length that are identical over the entirety of the shorter sequence are not
#     considered duplicates.

#     By default, comparison is made on the nucleotide sequence. To use the amino acid sequence instead,
#     set aa=True.
#     '''
#     nr_pairs = []
#     just_pairs = [p for p in pairs if p.is_pair]
#     single_chains = [p for p in pairs if not p.is_pair]
#     _pairs = just_pairs + single_chains
#     for p in _pairs:
#         duplicates = []
#         for nr in nr_pairs:
#             identical = True
#             vdj = 'vdj_aa' if aa else 'vdj_nt'
#             offset = 4 if aa else 12
#             if p.heavy is not None:
#                 if nr.heavy is None:
#                     identical = False
#                 else:
#                     heavy = p.heavy[vdj][offset:-offset] if ignore_primer_regions else p.heavy[vdj]
#                     nr_heavy = nr.heavy[vdj][offset:-offset] if ignore_primer_regions else nr.heavy[vdj]
#                     if heavy != nr_heavy:
#                         identical = False
#             if p.light is not None:
#                 if nr.light is None:
#                     identical = False
#                 else:
#                     light = p.light[vdj][offset:-offset] if ignore_primer_regions else p.light[vdj]
#                     nr_light = nr.light[vdj][offset:-offset] if ignore_primer_regions else nr.light[vdj]
#                     if light != nr_light:
#                         identical = False
#             duplicates.append(identical)
#         if any(duplicates):
#             continue
#         else:
#             nr_pairs.append(p)
#     return nr_pairs


# def refine(pairs, heavy=True, light=True, species='human'):
#     refined_pairs = copy.deepcopy(pairs)
#     for p in refined_pairs:
#         p.refine(heavy, light, species)
#     return refined_pairs


def vrc01_like(pair, loose=False):
    if loose:
        return loose_vrc01_like(pair)
    else:
        return strict_vrc01_like(pair)


def strict_vrc01_like(pair):
    if any([pair.heavy is None, pair.light is None]):
        vrc01_like = False
    else:
        try:
            # abstar's JSON output format
            vrc01_like = all(
                [pair.heavy["v_gene"]["gene"] == "IGHV1-2", pair.light["cdr3_len"] == 5]
            )
        except KeyError:
            try:
                # AIRR format
                vrc01_like = all(
                    [
                        pair.heavy["v_call"] == "IGHV1-2",
                        pair.light["junction_aa_length"] == 7,
                    ]
                )
            except KeyError:
                # unknown format
                vrc01_like = None
    return vrc01_like


def loose_vrc01_like(pair):
    try:
        # abstar's JSON output format
        if any([l["cdr3_len"] == 5 for l in pair.lights]) and any(
            [h["v_gene"]["gene"] == "IGHV1-2" for h in pair.heavies]
        ):
            loose_vrc01_like = True
        else:
            loose_vrc01_like = False
    except KeyError:
        try:
            # AIRR format
            if any([l["junction_aa_length"] == 7 for l in pair.lights]) and any(
                [h["v_call"] == "IGHV1-2" for h in pair.heavies]
            ):
                loose_vrc01_like = True
            else:
                loose_vrc01_like = False
        except:
            # unknown format
            loose_vrc01_like = None
    return loose_vrc01_like



#!/usr/bin/env python
# filename: alignment.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import multiprocessing as mp
import time
from contextlib import contextmanager
from typing import Iterable, Optional

from tqdm.auto import tqdm

from . import progbar

__all__ = ["monitor_mp_jobs", "monitor_celery_jobs"]


# def monitor_mp_jobs(
#     results: Iterable,
#     start_time: time.time = None,
#     completion_string: str = "\n",
#     print_progress: bool = True,
# ):
#     """
#     Monitors the progress of a set of multiprocessing jobs.

#     Parameters
#     ----------
#     results : Iterable[mp.pool.AsyncResult]
#         A list of AsyncResult objects from multiprocessing jobs.

#     start_time : time.time
#         The time at which the jobs started. If not provided, elapsed time will not be shown.

#     completion_string : str
#         A string to print when the jobs are complete. Default is a newline.

#     print_progress : bool
#         Whether or not to print a progress bar. Default is ``True``.

#     """
#     finished = 0
#     jobs = len(results)
#     while finished < jobs:
#         time.sleep(1)
#         ready = [ar for ar in results if ar.ready()]
#         finished = len(ready)
#         if print_progress:
#             progbar.progress_bar(finished, jobs, start_time=start_time)
#     if print_progress:
#         progbar.progress_bar(
#             finished,
#             jobs,
#             start_time=start_time,
#             complete=True,
#             completion_string=completion_string,
#         )


def monitor_mp_jobs(
    results: Iterable,
    completion_string: Optional[str] = None,
    print_progress: bool = True,
    start_time: Optional[time.time] = None,
):
    """
    Monitors the progress of a set of multiprocessing jobs.

    Parameters
    ----------
    results : Iterable[mp.pool.AsyncResult]
        A list of AsyncResult objects from multiprocessing jobs.

    completion_string : str, optional
        A string to print after the progress bar when the jobs are complete.

    print_progress : bool
        Whether or not to print a progress bar. Default is ``True``.

    start_time : time.time, optional
        DEPRECATED and is ignored. ``tqdm`` automatically handles the start time,
        but the argument is maintained for backwards compatibility.

    """
    jobs = len(results)
    finished = 0
    if print_progress:
        # initialize the TQDM progress bar
        pbar = tqdm(total=jobs)
        prev = 0
    # monitor job progress
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        if print_progress:
            update = finished - prev
            pbar.update(update)
            prev = finished
    if print_progress:
        pbar.close()
    if completion_string is not None:
        print(completion_string)


def monitor_celery_jobs(
    results: Iterable,
    start_time: time.time = None,
    completion_string: str = "\n",
    print_progress: bool = True,
):
    """
    Monitors the progress of a set of Celery jobs.

    Parameters
    ----------
    results : Iterable[celery.result.AsyncResult]
        A list of AsyncResult objects from Celery jobs.

    start_time : time.time
        The time at which the jobs started. If not provided, elapsed time will not be shown.

    completion_string : str
        A string to print when the jobs are complete. Default is a newline.

    print_progress : bool
        Whether or not to print a progress bar. Default is ``True``.

    """
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        succeeded = [ar for ar in results if ar.successful()]
        failed = [ar for ar in results if ar.failed()]
        finished = len(succeeded) + len(failed)
        if print_progress:
            progbar.progress_bar(finished, jobs, start_time=start_time)
    if print_progress:
        progbar.progress_bar(
            finished,
            jobs,
            start_time=start_time,
            complete=True,
            completion_string=completion_string,
        )


@contextmanager
def poolcontext(*args, **kwargs):
    pool = mp.Pool(*args, **kwargs)
    yield pool
    pool.terminate()



#!/usr/bin/env python
# filename: tl.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from .core.lineage import group_lineages
from .core.pair import assign_pairs
from .core.sequence import codon_optimize, reverse_complement, translate
from .tools.alignment import *
from .tools.anarci import *
from .tools.clonify import clonify
from .tools.cloning import *
from .tools.cluster import *
from .tools.phylo import *
from .tools.search import *
from .tools.similarity import repertoire_similarity
from .utils.phylogeny import igphyml, lsd

# from .utils.alignment import (
#     global_alignment,
#     local_alignment,
#     dot_alignment,
#     muscle,
#     muscle_v3,
#     mafft,
# )
# from .utils.cluster import cluster, cluster_vsearch, cluster_mmseqs
from .utils.progbar import progress_bar



#!/usr/bin/env python
# filename: pipeline.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# from __future__ import absolute_import, division, print_function, unicode_literals

import glob
import os
import sys
from typing import Iterable, Optional, Union

from . import log


# for backward compatibility
def list_files(*args, **kwargs):
    from ..io import list_files

    return list_files(*args, **kwargs)


def make_dir(*args, **kwargs):
    from ..io import make_dir

    return make_dir(*args, **kwargs)


def initialize(log_file, project_dir=None, debug=False):
    """
    Initializes an AbTools pipeline.

    Initialization includes printing the AbTools splash, setting up logging,
    creating the project directory, and logging both the project directory
    and the log location.

    Parameters
    ----------
    log_file : str
        Path to the log file. Required.

    project_dir : str
        Path to the project directory. If not provided,
        the project directory won't be created and the location won't be logged.

    debug : bool
        If ``True``, the logging level will be set to ``logging.DEBUG``.

    Returns
    -------
    logger
        A logger instance.

    """
    print_splash()
    log.setup_logging(log_file, print_log_location=False, debug=debug)
    logger = log.get_logger("pipeline")
    if project_dir is not None:
        make_dir(os.path.normpath(project_dir))
        logger.info("PROJECT DIRECTORY: {}".format(project_dir))
        logger.info("")
    logger.info("LOG LOCATION: {}".format(log_file))
    print("")
    return logger


# def make_dir(directory: str) -> None:
#     """
#     Makes a directory, if it doesn't already exist.

#     Parameters
#     ----------
#     directory : str
#         Path to a directory.

#     """
#     if not os.path.exists(directory):
#         os.makedirs(os.path.abspath(directory))


# def list_files(
#     directory: str, extension: Union[str, Iterable, None] = None
# ) -> Iterable[str]:
#     """
#     Lists files in a given directory.

#     Parameters
#     ----------
#     directory : str
#         Path to a directory.

#     extension : str
#         If supplied, only files that end with the specificied extension(s) will be returned. Can be either
#         a string or a list of strings. Extension evaluation is case-insensitive and can match complex
#         extensions (e.g. '.fastq.gz'). Default is ``None``, which returns all files in the directory,
#         regardless of extension.

#     Returns
#     -------
#     Iterable[str]

#     """
#     if os.path.isdir(directory):
#         expanded_dir = os.path.expanduser(directory)
#         files = sorted(glob.glob(expanded_dir + "/*"))
#     else:
#         files = [
#             directory,
#         ]
#     if extension is not None:
#         if isinstance(extension, str):
#             extension = [
#                 extension,
#             ]
#         files = [
#             f
#             for f in files
#             if any(
#                 [
#                     any([f.lower().endswith(e.lower()) for e in extension]),
#                     any([f.endswith(e.upper()) for e in extension]),
#                     any([f.endswith(e.lower()) for e in extension]),
#                 ]
#             )
#         ]
#     return files


def print_splash():
    splash = """
     _    _     _   _ _   _ _       ____  _            _ _            
    / \  | |__ | | | | |_(_) |___  |  _ \(_)_ __   ___| (_)_ __   ___ 
   / _ \ | '_ \| | | | __| | / __| | |_) | | '_ \ / _ \ | | '_ \ / _ \\
  / ___ \| |_) | |_| | |_| | \__ \ |  __/| | |_) |  __/ | | | | |  __/
 /_/   \_\_.__/ \___/ \__|_|_|___/ |_|   |_| .__/ \___|_|_|_| |_|\___|
                                           |_|                        
    """
    print("")
    print(splash)
    print(
        "(c) 2023 Bryan Briney\nDistributed under the MIT License (http://opensource.org/licenses/MIT)"
    )
    print("")
